package ca.ubc.cs.beta.hydra.builders;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.management.ManagementFactory;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.TreeMap;
import java.util.Map.Entry;

import ca.ubc.cs.beta.hydra.options.HydraOptions;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import com.beust.jcommander.JCommander;
import com.beust.jcommander.ParameterException;

import ca.ubc.cs.beta.aeatk.acquisitionfunctions.AcquisitionFunctions;
import ca.ubc.cs.beta.aeatk.algorithmexecutionconfiguration.AlgorithmExecutionConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunconfiguration.AlgorithmRunConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.AlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.ExistingAlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.RunStatus;
import ca.ubc.cs.beta.aeatk.exceptions.StateSerializationException;
import ca.ubc.cs.beta.aeatk.exceptions.TrajectoryDivergenceException;
import ca.ubc.cs.beta.aeatk.json.JSONConverter;
import ca.ubc.cs.beta.aeatk.misc.associatedvalue.Pair;
import ca.ubc.cs.beta.aeatk.misc.jcommander.JCommanderHelper;
import ca.ubc.cs.beta.aeatk.misc.returnvalues.AEATKReturnValues;
import ca.ubc.cs.beta.aeatk.misc.spi.SPIClassLoaderHelper;
import ca.ubc.cs.beta.aeatk.misc.version.VersionTracker;
import ca.ubc.cs.beta.aeatk.objectives.PortfolioObjective;
import ca.ubc.cs.beta.aeatk.objectives.RunObjective;
import ca.ubc.cs.beta.aeatk.options.AbstractOptions;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration.ParameterStringFormat;
import ca.ubc.cs.beta.aeatk.probleminstance.InstanceListWithSeeds;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstanceSeedPair;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstanceOptions.TrainTestInstances;
import ca.ubc.cs.beta.aeatk.random.SeedableRandomPool;
import ca.ubc.cs.beta.aeatk.runhistory.NewRunHistory;
import ca.ubc.cs.beta.aeatk.runhistory.RunHistory;
import ca.ubc.cs.beta.aeatk.smac.SMACOptions;
import ca.ubc.cs.beta.aeatk.state.StateFactoryOptions;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.base.cli.CommandLineAlgorithmRun;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.debug.SanityCheckingTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.functionality.OutstandingEvaluationsTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.functionality.portfolio.PortfolioRunKillingPolicy;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.functionality.portfolio.PortfolioTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.helpers.StrictlyIncreasingRuntimesTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.helpers.UseDynamicCappingExclusivelyTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.resource.caching.CachingTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.exceptions.TargetAlgorithmAbortException;
import ca.ubc.cs.beta.smac.builder.SMACBuilder;
import ca.ubc.cs.beta.smac.configurator.AbstractAlgorithmFramework;

public class SMACHydraWrapper {
    
    private static Logger log;
    private static Marker exception;
    private static Marker stackTrace;
    
    private static String logLocation = "<NO LOG LOCATION SPECIFIED, FAILURE MUST HAVE OCCURED EARLY>";
    
    private static InstanceListWithSeeds trainingILWS;
    private static InstanceListWithSeeds testingILWS;
    
    
    private static Map<String,  AbstractOptions> taeOptions;
    private static SeedableRandomPool pool;
    
    private static String outputDir;   
    
    private AlgorithmRunResult fRunResult;
    
    /**
     * Executes SMAC according to the given arguments
     * @param args  string input arguments
     * @return return value for operating system
     * @throws Throwable 
     */
    public SMACHydraWrapper(AlgorithmRunConfiguration runConfig) throws Throwable
    {  
        try {
            String aInstanceName = runConfig.getProblemInstanceSeedPair().getProblemInstance().getInstanceName();
            String aInstanceSpecificInformation = runConfig.getProblemInstanceSeedPair().getProblemInstance().getInstanceSpecificInformation();
            
            SMACInput input = SMACInput.fSMACInputConv.getObject(aInstanceSpecificInformation);
            
            List<AlgorithmRunResult> runs = input.fRuns;
            String[] args = input.fSMACArgs;
            List<ParameterConfiguration> runConfigs = input.fPortfolio;
            
//            //Randomly fail
//            Random rand = new Random();
//            if( rand.nextInt() % 5 == 0 )
//            {
//            	throw new IllegalStateException("Dummy Failure!");
//            }
//        
            
            final HydraOptions options = parseCLIOptions(args,aInstanceName);
            SMACOptions smacOptions = options.smacOptions;
            
            
            //RunHistory modelRH
            //List<Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>> fPortfolios
            
            final int origNumRun = smacOptions.seedOptions.numRun;
            /* Set seed that is unique to SMAC run */
            smacOptions.seedOptions.seedOffset = (int) runConfig.getProblemInstanceSeedPair().getSeed();
            final AlgorithmExecutionConfiguration execConfig = smacOptions.getAlgorithmExecutionConfig();
            final List<Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>> portfolio = new ArrayList<Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>>();

            TargetAlgorithmEvaluator tae = smacOptions.scenarioConfig.algoExecOptions.taeOpts.getTargetAlgorithmEvaluator( taeOptions, outputDir, smacOptions.seedOptions.numRun);
            
            /* ********** CAUTION! ******** 
             * Order of decorators matter
             */
            
            /*Warning: caching TAE may need a lot of RAM */
            

            
            /* Get Portfolio from parameter configurations */
            for(ParameterConfiguration config : runConfigs){
                portfolio.add(new Pair<AlgorithmExecutionConfiguration,ParameterConfiguration>(execConfig,config));
            }
            
            /* Cache previous runs. Should help manage portfolio execution */
            tae = new CachingTargetAlgorithmEvaluatorDecorator(tae); /* Apparently ExclusivelyTargetAlgorithmEvaluatorDecorator speeds up caching TAE */
            //tae = new CachingTargetAlgorithmEvaluatorDecorator(tae, runs, portfolio);
            
            /* Monitor runs that have not yet finished */
            tae = new OutstandingEvaluationsTargetAlgorithmEvaluatorDecorator(tae);
            /* Makes sure runtime is monotonically increasing for a run */
            tae = new StrictlyIncreasingRuntimesTargetAlgorithmEvaluatorDecorator(tae);
            /* Dynamically cap run at cutoff equal to runtime of incumbent */
            tae = new UseDynamicCappingExclusivelyTargetAlgorithmEvaluatorDecorator(tae);
            
            final NewRunHistory modelRH = new NewRunHistory(smacOptions.scenarioConfig.getIntraInstanceObjective(), smacOptions.scenarioConfig.interInstanceObj, smacOptions.scenarioConfig._runObj);
            
            

            
            if(portfolio.size() > 0)
            {
                if(portfolio.size() > 1)
                {
                    //TODO: Check that resubmit runs option is appropriately set
                    tae = new PortfolioTargetAlgorithmEvaluatorDecorator(tae, portfolio, smacOptions.scenarioConfig._runObj, true /*May want to change to true. Don't understand */, PortfolioRunKillingPolicy.SLOWERDIES, options.portfolioObjective);
                    /* Give SMAC hydra options for portfolio configuration */
                    options.smacOptions.challengerSelectInstances = options.challengerSelectInstances;
                }
                
                if(portfolio.size() == 1)
                {
                    tae = new PortfolioTargetAlgorithmEvaluatorDecorator(tae, portfolio, smacOptions.scenarioConfig._runObj, true, PortfolioRunKillingPolicy.SLOWERDIES, options.portfolioObjective);
                }
                
    			if (options.portfolioObjective.equals(PortfolioObjective.MARGINALIMPROVEMENT) || options.portfolioObjective.equals(PortfolioObjective.LOGMARGINALIMPROVEMENT ))
    			{
    				// Objective must be set to Quality for marginal improvement to support negative values
    				if(options.smacOptions.scenarioConfig._runObj.equals(RunObjective.RUNTIME))
    				{
        				options.smacOptions.expFunc = AcquisitionFunctions.EI;
        				options.smacOptions.scenarioConfig._runObj = RunObjective.QUALITY;
        				options.smacOptions.adaptiveCapping = true;
    				}
    				//Overwrite smac options for marginal improvement

    			}
            }
            /* See if TAE obeys post-conditions ?????*/
            tae = new SanityCheckingTargetAlgorithmEvaluatorDecorator(tae);
            
            /* Decorate TAE with unique identifier in instance specific information. We can do this since instances are uniquely identified by there name ( we can still properly compare instances) */
            if (options.overwriteInstanceSpecificInfoWithUniqueSMACIdentifier)
            {
            	tae = new SMACIdentifierTargetAlgorithmEvalautorDecorator(tae, Integer.toString(runConfig.getProblemInstanceSeedPair().getProblemInstance().getInstanceID()));
            }
            
            SMACBuilder smacBuilder = new SMACBuilder();
            
            AbstractAlgorithmFramework smac;
            
            String outputDirRun = outputDir;
            SeedableRandomPool fPool = smacOptions.seedOptions.getSeedableRandomPool();
            //TODO: Warm Start option to Hydra
            if (options.doWarmStart)
            {
                /* Get run history from run results */
                for(AlgorithmRunResult run : runs){
                    modelRH.append(run);
                }
                smac = smacBuilder.getAutomaticConfigurator(execConfig,  trainingILWS, smacOptions, taeOptions, outputDirRun, fPool, tae, modelRH);
            }else
            {
                smac = smacBuilder.getAutomaticConfigurator(execConfig,  trainingILWS, smacOptions, taeOptions, outputDirRun, fPool, tae, null);
            }
            smacOptions.shutdownTAEWhenDone = false;
            
            log.debug("Final TAE Built :{}", tae);
            smac.run();
            
            //smac.getModel();
            ParameterConfiguration newInc = smac.getIncumbent();
            //smac.
            
			JSONConverter<ParameterConfiguration> configConv = new JSONConverter<ParameterConfiguration>(){};
			String json = configConv.getJSON(newInc);
			log.info("New incumbent (JSON): {}", json);
            
            
            /* TODO: Make sure all instances have been run for the incumbents. Scrath that. We're doing lazy evaluation */ 
            
            List<AlgorithmRunResult> runResults = new ArrayList<AlgorithmRunResult>();
            for(AlgorithmRunResult run : smac.getRunHistory().getAlgorithmRunsExcludingRedundant())
            {
                runResults.add(run);
            }
            
            RunStatus aresult = RunStatus.SAT;
            double runtime = options.smacOptions.scenarioConfig.limitOptions.runtimeLimit;
            SMACResult  smacResult= new SMACResult(newInc,runResults,smac.getTerminationReason());
            String additionalRunData = SMACResult.fSMACResultConv.getJSON(smacResult);
            
            log.debug("Length of SMAC result JSON string: {} million characters", additionalRunData.length()/1000000.0);
            
            long aSeed = 0;
            fRunResult = new ExistingAlgorithmRunResult(runConfig,aresult,runtime, 0, -1, aSeed, additionalRunData);
            log.info("SMAC Execution Completed because: {}", smac.getTerminationReason());
        return;
    } catch(Throwable t)
    {
        System.out.flush();
        System.err.flush();
        
        System.err.println("Error occured running SMAC ( " + t.getClass().getSimpleName() + " : "+ t.getMessage() +  ")\nError Log: " + logLocation);
        System.err.flush();
        
        /* Write error to file */
        
        File f = new File(outputDir + File.separator + "error.txt");
        
        try{
            FileWriter fWrite = new FileWriter(f); 
            
            fWrite.append("Error occured running SMAC ( " + t.getClass().getSimpleName() + " : "+ t.getMessage() +  ")\nError Log: " + logLocation + "\n");
            fWrite.append(t.getMessage() + "\n");
            fWrite.append(t.getStackTrace().toString());
            
            fWrite.flush();
            fWrite.close();
        }catch (Exception e){
            
            System.err.println("Error writing error file. Reason: " + e.toString());        
        }
    
        
            if(log != null)
            {
                
                log.error( "Message: {}",t.getMessage());
                
                
                if(t instanceof ParameterException)
                {
                    log.info("Don't forget that some options are set by default from files in ~/.aclib/");
                    log.info("Exception stack trace", t);
                    
                    
                } else if(t instanceof TargetAlgorithmAbortException)
                {
                    log.error("A serious problem occured during target algorithm execution and we are aborting execution ",t );
                    
                    
                    log.error("We tried to call the target algorithm wrapper, but this call failed.");
                    log.error("The problem is (most likely) somewhere in the wrapper.");
                    log.error("There is also possibly additional error information above (in this log)");
                    log.error("The easiest way to debug this problem is to manually execute the call we tried and see why it did not return the correct result");
                    log.error("The required syntax is something like \"Final Result for ParamILS: x,x,x,x,x\".);");
                    log.error("Specifically the regex we are matching is {}", CommandLineAlgorithmRun.AUTOMATIC_CONFIGURATOR_RESULT_REGEX);
                }   else
                {
                    log.info("Maybe try running in DEBUG mode if you are missing information");
                    
                    log.error( "Exception:{}", t.getClass().getCanonicalName());
                    StringWriter sWriter = new StringWriter();
                    PrintWriter writer = new PrintWriter(sWriter);
                    t.printStackTrace(writer);
                    log.error( "StackTrace:{}",sWriter.toString());
                }
                    
                log.info("Exiting SMAC with failure. Log: " + logLocation);
                log.info("For a list of available commands use:  --help");
                log.info("Please see above for the available options. Further information is available in the following documents:");
                log.info("- The FAQ (doc/faq.pdf) contains commonly asked questions regarding troubleshooting, and usage.");
                log.info("- The Quickstart Guide (doc/quickstart.pdf) gives a simple example for getting up and running.");
                log.info("- The Manual (doc/manual.pdf) contains detailed information on file format semantics.");

                
                
                t = t.getCause();
            } else
            {
                if(t instanceof ParameterException )
                {
                    
                    System.err.println(t.getMessage());
                    t.printStackTrace();
                } else
                {
                    t.printStackTrace();
                }
                
            }
    
            //TODO: Get proper Exceptions:
            /*
            if(t instanceof ParameterException)
            {
                return AEATKReturnValues.PARAMETER_EXCEPTION;
            }
            
            if(t instanceof StateSerializationException)
            {
                return AEATKReturnValues.SERIALIZATION_EXCEPTION;
            }
            
            if(t instanceof TrajectoryDivergenceException)
            {
                return AEATKReturnValues.TRAJECTORY_DIVERGENCE;
            }
            
            return AEATKReturnValues.OTHER_EXCEPTION;
            */
            throw t;
    }
    
    
}



private static String runGroupName = "DEFAULT";

/**
 * Parsers Command Line Arguments and returns a options object
 * @param args
 * @return
 */
private static HydraOptions parseCLIOptions(String[] args, String instanceName) throws ParameterException, IOException
{
    //DO NOT LOG UNTIL AFTER WE PARSE CONFIG OBJECT
    
    
    /* What about hydra parameters???*/
    
    HydraOptions options = new HydraOptions();
    
    taeOptions = options.smacOptions.scenarioConfig.algoExecOptions.taeOpts.getAvailableTargetAlgorithmEvaluators();
    JCommander jcom = JCommanderHelper.getJCommanderAndCheckForHelp(args, options, taeOptions);
    
    jcom.setProgramName("hydra");
    
    try {
        try {
            try {
            
            
            args = processScenarioStateRestore(args);
            jcom.parse(args);
            } finally
            {
                runGroupName = options.smacOptions.runGroupOptions.getFailbackRunGroup();
            }
            
            

            if(options.smacOptions.adaptiveCapping == null)
            {
                switch(options.smacOptions.scenarioConfig._runObj)
                {
                case RUNTIME:
                    options.smacOptions.adaptiveCapping = true;
                    break;
                    
                case QUALITY:
                    options.smacOptions.adaptiveCapping = false;
                    break;
                    
                default:
                    //You need to add something new here
                    throw new IllegalStateException("Not sure what to default too");
                }
            }
            
            if(options.smacOptions.randomForestOptions.logModel == null)
            {
                switch(options.smacOptions.scenarioConfig._runObj)
                {
                case RUNTIME:
                    options.smacOptions.randomForestOptions.logModel = true;
                    break;
                case QUALITY:
                    options.smacOptions.randomForestOptions.logModel = false;
                }
            }
            runGroupName = options.smacOptions.getRunGroupName(taeOptions.values());
            //File outputDir = new File(options.scenarioConfig.outputDirectory);
            
            
            /*
             * TODO: Think about directory structure
             * Build the Serializer object used in the model 
             */
            //outputDir = options.getSMACOutputDirectory(runGroupName);
            outputDir = options.getSMACOutputDirectory(instanceName);
            
            File outputDirFile = new File(outputDir);
            
            
            if(!outputDirFile.exists())
            {
                outputDirFile.mkdirs();
                //Check again to ensure there isn't a race condition
                if(!outputDirFile.exists())
                {
                    throw new ParameterException("Could not create all folders necessary for output directory: " + outputDir);
                }
                
            }
            
            
        } finally
        {
            
            options.smacOptions.logOptions.initializeLogging(outputDir, options.smacOptions.seedOptions.numRun);
            SMACHydraWrapper.logLocation = options.smacOptions.logOptions.getLogLocation(outputDir,options.smacOptions.seedOptions.numRun);
            
            log = LoggerFactory.getLogger(SMACHydraWrapper.class);
            
            //log.ROOT_LOGGER_NAME;
            
            exception = MarkerFactory.getMarker("EXCEPTION");
            stackTrace = MarkerFactory.getMarker("STACKTRACE");
            
            VersionTracker.setClassLoader(SPIClassLoaderHelper.getClassLoader());
            VersionTracker.logVersions();
            
            for(String name : jcom.getParameterFilesToRead())
            {
                log.info("Parsing (default) options from file: {} ", name);
            }
            
            log.info("Logging to {}." , outputDir);
        }
        
        log.trace("Command Line Options Parsed");
        

        
        JCommanderHelper.logCallString(args, SMACHydraWrapper.class);
        

         if(log.isDebugEnabled())
         {
            Map<String, String> env = new TreeMap<String, String>(System.getenv());
                
            StringBuilder sb = new StringBuilder();
             for (String envName : env.keySet()) {
                 sb.append(envName).append("=").append(env.get(envName)).append("\n");
                 
                   
                }
            
            
                 
             log.debug("==========Enviroment Variables===========\n{}", sb.toString());
             
             
             Map<Object,Object > props = new TreeMap<Object, Object>(System.getProperties());
             sb = new StringBuilder();
             for (Entry<Object, Object> ent : props.entrySet())
             {
                 
                 sb.append(ent.getKey().toString()).append("=").append(ent.getValue().toString()).append("\n");  
             }
            
             String hostname = "[UNABLE TO DETERMINE HOSTNAME]";
            try {
                hostname = InetAddress.getLocalHost().getHostName();
            } catch(UnknownHostException e)
            { //If this fails it's okay we just use it to output to the log
                
            }
            
            log.debug("Hostname:{}", hostname);
            log.debug("==========System Properties==============\n{}", sb.toString() );
         }
        
        JCommanderHelper.logConfiguration(jcom);
        pool = options.smacOptions.seedOptions.getSeedableRandomPool();
        

        TrainTestInstances tti = options.smacOptions.getTrainingAndTestProblemInstances(pool, new SeedableRandomPool(options.smacOptions.validationSeed + options.smacOptions.seedOptions.seedOffset,pool.getInitialSeeds()));
        trainingILWS = tti.getTrainingInstances();
        testingILWS = tti.getTestInstances();
    
        try {
            //We don't handle this more gracefully because this seems like a super rare incident.
            if(ManagementFactory.getThreadMXBean().isThreadCpuTimeEnabled())
            {
                log.debug("JVM Supports CPU Timing Measurements");
            } else
            {
                log.warn("This Java Virtual Machine has CPU Time Measurements disabled, tunerTimeout will not contain any SMAC Execution Time.");
            }
        } catch(UnsupportedOperationException e)
        {
            log.warn("This Java Virtual Machine does not support CPU Time Measurements, tunerTimeout will not contain any SMAC Execution Time Information (http://docs.oracle.com/javase/1.5.0/docs/api/java/lang/management/ThreadMXBean.html#setThreadCpuTimeEnabled(boolean))");
        }
        
        if(options.smacOptions.seedOptions.numRun + options.smacOptions.seedOptions.seedOffset < 0)
        {
            log.warn("NumRun {} plus Seed Offset {} should be positive, things may not seed correctly",options.smacOptions.seedOptions.numRun, options.smacOptions.seedOptions.seedOffset );
        }
        
        return options;
    } catch(IOException e)
    {
        throw e;
        
    } catch(ParameterException e)
    {
        throw e;
    }
}

private static String[] processScenarioStateRestore(String[] args) {
    return StateFactoryOptions.processScenarioStateRestore(args);

}
    



public AlgorithmRunResult getResult() {
    return fRunResult;
}
    


            
            

}
