package ca.ubc.cs.beta.hydra.builders;

import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ca.ubc.cs.beta.aeatk.algorithmexecutionconfiguration.AlgorithmExecutionConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunconfiguration.AlgorithmRunConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.AlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstanceSeedPair;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluatorCallback;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluatorRunObserver;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.AbstractTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.zilla.algorithmexecution.managers.FromFileAlgorithmExecutionManager;

public class SMACIdentifierTargetAlgorithmEvalautorDecorator extends AbstractTargetAlgorithmEvaluatorDecorator {
	
	private String fIdentifier;
	
	private static Logger log = LoggerFactory
			.getLogger(SMACIdentifierTargetAlgorithmEvalautorDecorator.class);

	public SMACIdentifierTargetAlgorithmEvalautorDecorator(TargetAlgorithmEvaluator tae, String identifier)
	{
		super(tae);
		this.fIdentifier = identifier;
		log.info("Decorating TAE that allows unique idneitifaction of SMAC..");

	}
	
	@Override
	public final List<AlgorithmRunResult> evaluateRun(List<AlgorithmRunConfiguration> runConfigs, TargetAlgorithmEvaluatorRunObserver obs) {
		
		log.trace("Intitial Run Configs: {}", runConfigs);
		List<AlgorithmRunConfiguration> modifiedRunConfigs = new ArrayList<AlgorithmRunConfiguration>();
		for (AlgorithmRunConfiguration runConfig : runConfigs)
		{
			modifiedRunConfigs.add(addIdentifier(runConfig));
		}
		log.trace("Modified Run Configs: {}", modifiedRunConfigs);
		return tae.evaluateRun(modifiedRunConfigs, obs);
		
	}

	@Override
	public final void evaluateRunsAsync(final List<AlgorithmRunConfiguration> runConfigs, final TargetAlgorithmEvaluatorCallback oHandler, TargetAlgorithmEvaluatorRunObserver obs)
	{
		List<AlgorithmRunConfiguration> modifiedRunConfigs = new ArrayList<AlgorithmRunConfiguration>();
		for (AlgorithmRunConfiguration runConfig : runConfigs)
		{
			modifiedRunConfigs.add(addIdentifier(runConfig));
		}
		tae.evaluateRunsAsync(modifiedRunConfigs, oHandler, obs);
	}

	private AlgorithmRunConfiguration addIdentifier(
			AlgorithmRunConfiguration runConfig) {
		
		ParameterConfiguration config = runConfig.getParameterConfiguration();
		AlgorithmExecutionConfiguration execConfig = runConfig.getAlgorithmExecutionConfiguration();
		ProblemInstanceSeedPair pisp= runConfig.getProblemInstanceSeedPair();
		ProblemInstance pi = pisp.getProblemInstance();
		long seed = pisp.getSeed();
		ProblemInstance modifiedProblemInstance = new ProblemInstance(pi.getInstanceName(),pi.getInstanceID(),pi.getFeatures(),fIdentifier);
		pisp = new ProblemInstanceSeedPair(modifiedProblemInstance, seed);
		AlgorithmRunConfiguration modifiedRunConfig = new AlgorithmRunConfiguration(pisp, config, execConfig);
		log.trace("RunConfig: {}, ModifiedRunConfig: {}", runConfig, modifiedRunConfig);
		
		return modifiedRunConfig;
	}

	@Override
	protected void postDecorateeNotifyShutdown() {
		// TODO Auto-generated method stub
		
	}


}
