package ca.ubc.cs.beta.hydra.executors;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.io.StringWriter;
import java.lang.management.ManagementFactory;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.Map.Entry;
import java.util.TreeSet;

import javax.imageio.ImageIO;
import javax.swing.JFrame;

import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.functionality.OutstandingEvaluationsTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.hydra.builders.CompressionTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.hydra.builders.HydraSMACBuilder;
import ca.ubc.cs.beta.hydra.builders.SMACInput;
import ca.ubc.cs.beta.hydra.builders.SMACResult;
import ca.ubc.cs.beta.hydra.incumbent.AllRunsIncumbentEvaluator;
import ca.ubc.cs.beta.hydra.incumbent.IIncumbentEvaluator;
import ca.ubc.cs.beta.hydra.incumbent.ModelIncumbentEvaluator;
import ca.ubc.cs.beta.hydra.options.HydraOptions;
import ca.ubc.cs.beta.hydra.portfoliohistory.IPortfolioEvaluator;
import ca.ubc.cs.beta.hydra.portfoliohistory.PortfolioHistory;
import ca.ubc.cs.beta.hydra.portfoliohistory.VBSPortfolioEvaluator;
import ca.ubc.cs.beta.hydra.portfoliohistory.ZillaPortfolioEvaluator;
import ca.ubc.cs.beta.zilla.algorithmexecution.AlgorithmExecutionResult;
import ca.ubc.cs.beta.zilla.algorithmexecution.managers.FromFileAlgorithmExecutionManager;
import ca.ubc.cs.beta.zilla.base.algorithms.Algorithm;
import ca.ubc.cs.beta.zilla.base.probleminstances.ProblemInstance;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;
//import org.math.plot.*;
//import org.math.plot.components.LegendPanel;
















import com.beust.jcommander.JCommander;
import com.beust.jcommander.ParameterException;

import ca.ubc.cs.beta.aeatk.exceptions.DuplicateRunException;
import ca.ubc.cs.beta.aeatk.exceptions.StateSerializationException;
import ca.ubc.cs.beta.aeatk.exceptions.TrajectoryDivergenceException;
import ca.ubc.cs.beta.aeatk.json.JSONConverter;
import ca.ubc.cs.beta.aeatk.misc.associatedvalue.Pair;
import ca.ubc.cs.beta.aeatk.misc.jcommander.JCommanderHelper;
import ca.ubc.cs.beta.aeatk.misc.returnvalues.AEATKReturnValues;
import ca.ubc.cs.beta.aeatk.misc.spi.SPIClassLoaderHelper;
import ca.ubc.cs.beta.aeatk.misc.version.VersionTracker;
import ca.ubc.cs.beta.aeatk.objectives.PortfolioObjective;
import ca.ubc.cs.beta.aeatk.objectives.RunObjective;
import ca.ubc.cs.beta.aeatk.options.AbstractOptions;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration.ParameterStringFormat;
import ca.ubc.cs.beta.aeatk.probleminstance.InstanceListWithSeeds;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstanceSeedPair;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstanceOptions.TrainTestInstances;
import ca.ubc.cs.beta.aeatk.random.SeedableRandomPool;
import ca.ubc.cs.beta.aeatk.runhistory.NewRunHistory;
import ca.ubc.cs.beta.aeatk.smac.SMACOptions;
import ca.ubc.cs.beta.aeatk.state.StateFactoryOptions;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.base.cli.CommandLineAlgorithmRun;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.helpers.StrictlyIncreasingRuntimesTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.helpers.UseDynamicCappingExclusivelyTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.resource.caching.CachingTargetAlgorithmEvaluatorDecorator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.exceptions.TargetAlgorithmAbortException;
import ca.ubc.cs.beta.aeatk.algorithmexecutionconfiguration.AlgorithmExecutionConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunconfiguration.AlgorithmRunConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.AlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.ExistingAlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.RunStatus;

/**
 * Executes Hydra (http://www.cs.ubc.ca/labs/beta/Projects/Hydra/)
 * Hydra takes as input a configurable algorithm, instance set, and parameter set. 
 * It returns a set of algorithm configurations to use for portfolio-based algorithm selection. 
 * Hydra is a greedy algorithm and proceeds by first determining a high-performance parameter configuration.
 * Then iteratively finds and add new configurations to the portfolio that improve portfolio-based selection 
 * performance.
 * @author Chris Cameron, Alexandre Frechette, Steve Ramage
 *
 */
public class HydraExecutor {

	
	private static Logger log;
	private static Marker exception;
	private static Marker stackTrace;
	
	public static String logLocation = "<NO LOG LOCATION SPECIFIED, FAILURE MUST HAVE OCCURED EARLY>";
	
	private static InstanceListWithSeeds trainingILWS;
	private static InstanceListWithSeeds testingILWS;
	
	
	private static Map<String,  AbstractOptions> taeOptions;
	private static SeedableRandomPool pool;
	
	private static String outputDir;
    private static ParameterConfiguration bestSingleConfiguration;
    private static double bestSingleConfigurationPerformance;
    
    private static AlgorithmExecutionConfiguration execConfig;
	private static NewRunHistory notFreeRH;
	
	/**
	 * Executes Hydra then exits the JVM {@see System.exit()}
	 *  
	 * @param args string arguments
	 */
	public static void main(String[] args)
	{
		int returnValue = hydraExecutor(args);
		
		if(log != null)
		{
			log.info("Returning with value: {}",returnValue);
		}
		
		System.exit(returnValue);
	}
	
	
	/**
	 * Executes SMAC according to the given arguments
	 * @param args 	string input arguments
	 * @return return value for operating system
	 */
	public static int hydraExecutor(String[] args)
	{
		/*
		 * WARNING: DO NOT LOG ANYTHING UNTIL AFTER WE HAVE PARSED THE CLI OPTIONS
		 * AS THE CLI OPTIONS USE A TRICK TO ALLOW LOGGING TO BE CONFIGURABLE ON THE CLI
		 * IF YOU LOG PRIOR TO IT ACTIVATING, IT WILL BE IGNORED 
		 */
		
		
		try {
			final HydraOptions options = parseCLIOptions(args);
			
			final SMACOptions smacOptions = options.smacOptions;
			
			final List<ParameterConfiguration> portfolio = new ArrayList<ParameterConfiguration>();

			final AlgorithmExecutionConfiguration execConfig = smacOptions.getAlgorithmExecutionConfig();
			
			
			TargetAlgorithmEvaluator myTAE = smacOptions.scenarioConfig.algoExecOptions.taeOpts.getTargetAlgorithmEvaluator( taeOptions, outputDir, smacOptions.seedOptions.numRun);
			
			/* ********** CAUTION! ******** 
			 * Order of decorators matter!
			 */

			myTAE = new CachingTargetAlgorithmEvaluatorDecorator(myTAE); /* Apparently ExclusivelyTargetAlgorithmEvaluatorDecorator speeds up caching TAE */
			/* Monitor runs that have not yet finished */
			myTAE = new OutstandingEvaluationsTargetAlgorithmEvaluatorDecorator(myTAE);
			/* Makes sure runtime is monotonically increasing for a run */
			myTAE = new StrictlyIncreasingRuntimesTargetAlgorithmEvaluatorDecorator(myTAE);
			/* Dynamically cap run at cutoff equal to runtime of incumbent */
			myTAE = new UseDynamicCappingExclusivelyTargetAlgorithmEvaluatorDecorator(myTAE);
			
			final NewRunHistory rh = new NewRunHistory(smacOptions.scenarioConfig.getIntraInstanceObjective(), smacOptions.scenarioConfig.interInstanceObj, smacOptions.scenarioConfig._runObj);
			
			smacOptions.shutdownTAEWhenDone = false;

			/* Get TAE for running SMAC executions */
		    TargetAlgorithmEvaluator smacExecutionsTAE = options.smacTAEOptions.getTargetAlgorithmEvaluator( taeOptions, outputDir, smacOptions.seedOptions.numRun);
	        /* Compress input to smac executions TAE (Required with SMACTAE) */
		    smacExecutionsTAE = new CompressionTargetAlgorithmEvaluatorDecorator(smacExecutionsTAE);
		    
		    
		    /* Keep track of hydra performance at each iteration */
		    Map<Integer, Double> hydraPerformance = new HashMap<Integer, Double>();
		    
		    /* How to evaluate incumbent configurations */
		    IIncumbentEvaluator incumbentEvaluator = null;
		    switch(options.incumbentEvaluation)
		    {
		    case SMAC_RUNS:
		        incumbentEvaluator = new AllRunsIncumbentEvaluator(smacOptions.scenarioConfig._runObj);
		        break;
		    /* TODO
		    case ALL_TRAINING:
		        incumbentEvaluator = new AllRunsIncumbentEvaluator(smacOptions.scenarioConfig._runObj);
		        break;
		    */
		    /* TODO
		    case SMAC_MODEL:
		        incumbentEvaluator = new ModelIncumbentEvaluator();
		        break;
		    */
		    }
		    
		    /* How to evaluate portfolio performance */
		    IPortfolioEvaluator portfolioEvaluator = null;
		    double testCutoff = options.getTestCutoff();
		    log.info("Running Hydra with test cutoff: " + testCutoff);
		    //RunObjective ro = options.smacOptions.scenarioConfig._runObj;
		    //RunObject
		    switch(options.portfolioEvaluation)
		    {
		    case ZILLA:
		        String dirName = outputDir + File.separator + "zilla";
		        File zillaDir = new File(outputDir + File.separator + "zilla");
		        zillaDir.mkdirs();
		        portfolioEvaluator = new ZillaPortfolioEvaluator(testCutoff, options.zillaOptions, smacOptions.seedOptions.getSeedableRandomPool(),myTAE,smacOptions.scenarioConfig._runObj, dirName, trainingILWS.getInstances(), testingILWS.getInstances(), options.smacOptions.scenarioConfig.instanceOptions.instanceFeatureFile, options.smacOptions.scenarioConfig.interInstanceObj);
		        break;
		    case VBS:
		        if(testingILWS.getInstances().isEmpty())
		        {
		            log.warn("Testing on training set since testing set is empty.");
		            portfolioEvaluator = new VBSPortfolioEvaluator(trainingILWS,myTAE, smacOptions.scenarioConfig._runObj, smacOptions.scenarioConfig.getIntraInstanceObjective(), testCutoff, options.portfolioObjective);
		        }else
		        {
		            portfolioEvaluator = new VBSPortfolioEvaluator(testingILWS,myTAE, smacOptions.scenarioConfig._runObj, smacOptions.scenarioConfig.getIntraInstanceObjective(),  testCutoff, options.portfolioObjective);
		        }
		        break;
		    }
		    
		    /* Create SMAC builder */
            HydraSMACBuilder hydraSMACBuilder = new HydraSMACBuilder(options,incumbentEvaluator);
		    
		    int initialIteration = 0;
		    
		    /****** Restarting *******/
		    if(options.restartFolder != null)
		    {
		        log.info("Restarting from: {}", options.restartFolder);
		        
		        File portfolioDir = new File(options.restartFolder + File.separator + "Portfolios");
		        log.info("Portfolio Directory: {}", portfolioDir.toString());
		        
		        /* Find the most recent iteration */
		        int mostRecentIteration = 0;
		        for(File f: portfolioDir.listFiles())
		        {
		            int iteration = Integer.parseInt(f.getName().split("-")[1].split(".txt")[0]);
		            hydraPerformance.put(iteration, 0.0);
		            if(iteration > mostRecentIteration);
		            {
		                mostRecentIteration = iteration;
		            }
		        }
		        initialIteration = mostRecentIteration+1;
		        log.info("Starting Hydra at iteration {}", initialIteration);

		        /* Parsing portfolio file and adding existing parameter configurations to portfolio */
		        File portfolioFile = new File(options.restartFolder + File.separator + "Portfolios" + File.separator + "portfolio-" + mostRecentIteration + ".txt");
		        JSONConverter<ParameterConfiguration> configConv = new JSONConverter<ParameterConfiguration>(){};
		        BufferedReader br = new BufferedReader(new FileReader(portfolioFile));  
		        String line = null;  
		        while ((line = br.readLine()) != null)  
		        {  
		            ParameterConfiguration config = configConv.getObject(line);
		            
		            /* Gather the best single configuration */
		            if (portfolio.size() == 0)
		            {
		                bestSingleConfiguration = config;
		                List<ParameterConfiguration> singlePortfolio = new ArrayList<ParameterConfiguration>();
		                singlePortfolio.add(bestSingleConfiguration);
		                bestSingleConfigurationPerformance = 50; //portfolioEvaluator.getPerformance(singlePortfolio,execConfig);
		            }

		            portfolio.add(config);
		        } 
		    }
		   
		   /* Keep track of newly added configuration to portfolio */
		   Set<ParameterConfiguration>  newlyAddedConfigs = new HashSet<ParameterConfiguration>();
		   int numIterationsExecuted = 0;
			try 
			{
			    log.info("Running Hydra for {} iterations", options.numIterations);
			    
				for(int iter=initialIteration; iter < options.numIterations; iter++)
				{
                    try
					{
					    log.info("**************Hydra iteration: {}***************", iter);
					
					    /* Get updated performance for portfolio of all algorithm execution configurations in run history
					     * Portfolio Performance of Config X = min(Portfolio Performance, Performance of Config X) 
					     */

                        
                        /* Information needed to run SMAC */
						/* Only pass runs of portfolio configs */
						List<AlgorithmRunResult> runsToSubmit = new ArrayList<AlgorithmRunResult>();
						if(!options.doWarmStart)
						{
							if (iter!=0)
							{
								log.warn("Warm starting is turned off. SMAC models will be built from scratch.");
							}
						    /* Unneccessary
						    for(AlgorithmRunResult result : modelRH.getAlgorithmRunsExcludingRedundant())
						    {
						        ParameterConfiguration config = result.getParameterConfiguration();
						        
                                if(newlyAddedConfigs.contains(config))
						        {
						            runsToSubmit.add(result);
						        }
						    }
						    */
						}else
						{
		                   PortfolioHistory portfolioHistory= new PortfolioHistory(rh, notFreeRH,smacOptions, myTAE, portfolio, newlyAddedConfigs,execConfig, options.portfolioObjective);
		                   NewRunHistory modelRH = portfolioHistory.getHistory();
						   runsToSubmit = modelRH.getAlgorithmRunsExcludingRedundant();
						   
						   log.debug("Number of runs to pass to SMAC reduced from {} to {}.", modelRH.getAlgorithmRunsExcludingRedundant().size(),runsToSubmit.size());
						   
//						   ByteArrayOutputStream baos = new ByteArrayOutputStream();
//						   ObjectOutputStream oos = new ObjectOutputStream(baos);
//						   oos.writeObject(modelRH);
//						   oos.close();
//						   double megaBytes = baos.size() / (1000.0*1000.0);
//						   log.info("Size of history: approximately {} MB", megaBytes);
						}
						

						
						
						/* Reset newly added Configs */
						newlyAddedConfigs = new HashSet<ParameterConfiguration>();
						
                        /* Create object containing SMAC input information */
                        SMACInput input = new SMACInput(args,runsToSubmit,portfolio);
                    
                        /* Create SMAC builder */
                        hydraSMACBuilder.buildSMACs(input,smacExecutionsTAE,iter, pool.getRandom("HydraSMACBuilder").nextLong());
						
                        /* Gather run results and new incumbents */
						List<AlgorithmRunResult> runResults = hydraSMACBuilder.getRunResults();
						List<ParameterConfiguration> newConfigurations = hydraSMACBuilder.getBestIncumbents();
						log.info("Found best incumbents.");
						
						/* Gather the best single configuration */
						if (portfolio.size() == 0)
						{
						    log.info("Setting Best Single configuration. Only occurs in first Hydra iteration");
						    bestSingleConfiguration = newConfigurations.get(0);

						    log.info("Best single configuration: {}", bestSingleConfiguration.getFormattedParameterString(ParameterStringFormat.NODB_SYNTAX));
						}
						
						pool.logUsage();
						
						/* Only keep track of run history if we are warm starting */
						if (options.doWarmStart)
						{
							notFreeRH = new NewRunHistory(smacOptions.scenarioConfig.getIntraInstanceObjective(), smacOptions.scenarioConfig.interInstanceObj, smacOptions.scenarioConfig._runObj);
							if (!options.addAllAlgorithmRuns)
							{
								/* Get instances from best SMAC run and add all run with common instances from other smac runs */
								int runHistorySeed = 0;
								ParameterConfiguration bestConfig = newConfigurations.get(0);
								List<ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance> commonInstances = new ArrayList<ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance>();
								for(AlgorithmRunResult run : runResults)
								{
									if (run.getParameterConfiguration().equals(bestConfig))
									{
										commonInstances.add(run.getProblemInstance());
									}

								}
								for(AlgorithmRunResult run : runResults)
								{
									if (commonInstances.contains(run.getProblemInstance()))
									{
										try
										{
											rh.append(run);
										} catch(DuplicateRunException e)
										{
											//Duplicate run who cares
										}
									}
									else
									{
										try {
											notFreeRH.append(run);
										} catch (DuplicateRunException e) {
											//Duplicate run who cares
										}
									}
								}
							}
							else
							{
								/* Add all run results */
								for(AlgorithmRunResult run : runResults)
								{
									try {
										log.debug("Algorithm execution configuration is {}",run.getAlgorithmExecutionConfiguration());
										rh.append(run);
									} catch(DuplicateRunException e)
									{
										//Duplicate run who cares
									}
									catch(IllegalArgumentException e)
									{
										throw new IllegalArgumentException(e);
									}
								}
							}
						}
					
						/* Add new configuration to portfolio */
				        boolean uniqueConfigFound = false;
				        log.info("Adding new incumbent configurations...");
						for(ParameterConfiguration config : newConfigurations)
						{
						    if(portfolio.contains(config))
						    {
						        log.info("SMAC returned an existing configuration in the portfolio: {}", config.getFormattedParameterString(ParameterStringFormat.NODB_SYNTAX));
						        break;
						    } else
						    {
						        log.info("Adding configuration: {}" , config.getFormattedParameterString(ParameterStringFormat.NODB_SYNTAX) );
						        portfolio.add(config);
						        newlyAddedConfigs.add(config);
						        uniqueConfigFound = true;
						    }
						}
						
						if(!uniqueConfigFound)
						{
						    log.info("HYDRA detected that all SMAC configurations are contained in the portfolio, this means that HYDRA has stagnated at this point");
						    numIterationsExecuted = iter+1;
						    break;
						}
						
						/* TODO: There is another termination condition if there is not significant improvement.
						 * What does not signficant improvement refer to. Maybe do regularized VBS subset selection??
						 */
						
						double performance = portfolioEvaluator.getPerformance(portfolio, execConfig, iter);
                        hydraPerformance.put(iter, performance);


                        /********* Evalalute Portfolio Performance and write to file ********/
                        
                        log.info("Test performance of hydra after iteration {}: {}", iter, performance);
                        if(hydraPerformance.keySet().contains(iter-1))
                        {
                            double previousIterationPerformance = hydraPerformance.get(iter-1);
                            int percentImprovement = (int) (100 * (previousIterationPerformance - performance) / Math.abs(previousIterationPerformance));
                            if(percentImprovement < 0)
                            {
                                switch (options.portfolioEvaluation)
                                {
                                    case VBS:
                                        log.warn("VBS performance has got worse! This must be a bug!");
                                    case ZILLA:
                                        log.warn("Zilla performance has got worse! Perhaps unpredictive configurations were added to portfolio.");
                                }
                            }
                            log.info("Test performance reduction of {}% over previous iteration", percentImprovement);
                        }
                        
						File f = new File(outputDir+File.separator + "Portfolios");
						f.mkdirs();
						f = new File(outputDir+File.separator + "Portfolios"+ File.separator+ "portfolio-"+iter+".txt");
						log.info("Writing current portfolio to: {}" ,f.toString());
						
						FileWriter fWrite = new FileWriter(f); 
						int j=0;
						JSONConverter<ParameterConfiguration> configConv = new JSONConverter<ParameterConfiguration>(){};
						for(ParameterConfiguration config : portfolio)
						{
						    String json = configConv.getJSON(config);
							fWrite.append(json + "\n");
							j++;
						}
						fWrite.flush();
						fWrite.close();
						
						
						if (options.savePortfolioPerInstancePerformance)
						{
							List<AlgorithmRunConfiguration> runConfigs = new ArrayList<AlgorithmRunConfiguration>();
				            for(ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance problem : testingILWS.getInstances())
				            {
				                for(ParameterConfiguration paramConfig : portfolio)
				                {
				                    /* Not sure if this is correct (instance seed pairs are correct). Might want to check cache */
				                    /* TODO: Can run out of seeds for an instance. What does this mean? */
				                    //long seed = fInstances.getSeedGen().getNextSeed(problem);
				                    long seed = 0;
				                    
				                    ProblemInstanceSeedPair pisp = new ProblemInstanceSeedPair(problem,seed);
				                    
									runConfigs.add(new AlgorithmRunConfiguration(pisp, paramConfig, execConfig));
				                }
				            }
				            
				            List<AlgorithmRunResult> portfolioRunResults = myTAE.evaluateRun(runConfigs);
				            
				            Map<Pair<ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance, ParameterConfiguration>,Double> portfolioRunResultsMap = new HashMap<Pair<ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance, ParameterConfiguration>,Double>();
				            
				            for (AlgorithmRunResult runResult : portfolioRunResults)
				            {
				            	Pair<ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance,ParameterConfiguration> key = new Pair<ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance, ParameterConfiguration>(runResult.getProblemInstance(),runResult.getParameterConfiguration());
				                Double value = options.smacOptions.scenarioConfig._runObj.getObjective(runResult);
				                portfolioRunResultsMap.put(key, value);
				            }
				            
				            /* Write File */
				            File portfolioResultsFile = new File(outputDir+File.separator + "portfolio_performance_iteration_" + iter + ".csv");
				            BufferedWriter bw = new BufferedWriter(new FileWriter(portfolioResultsFile));
				            List<String> header = new ArrayList<String>();
				            for(ParameterConfiguration config: portfolio)
				            {
				            	header.add("Config" + config.getFriendlyID());
				            }            
				            bw.write(StringUtils.join(header,",")+ System.lineSeparator());
				            
				            for (ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance problem : testingILWS.getInstances())
				            {
				            	List<Double> line = new ArrayList<Double>();
				                for(ParameterConfiguration paramConfig : portfolio)
				                {
				                	line.add(portfolioRunResultsMap.get(new Pair<ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance, ParameterConfiguration>(problem, paramConfig)));
				                }
				                String instanceEntry = StringUtils.join(line, ",") + System.lineSeparator();
				                
				                bw.write(instanceEntry);
				            }
				            bw.close();
				            
				            
						}


					} catch(Exception e)
					{
					    e.printStackTrace();
					    throw new RuntimeException(e);
					}finally
					{
						//Dispatch testing phase in different thread
						/*
                        class MyRunnable implements Runnable { 
                            public void run() { 
                                while ( true ) { 
                             	   
            					    double performance = portfolioEvaluator.getPerformance(portfolio, execConfig, iter);
            					    hydraPerformance.put(iter, performance);
                             	   
                             	   
                                    log.info("Test performance of hydra after iteration {}: {}", iter, performance);
                                    if(hydraPerformance.keySet().contains(iter-1))
                                    {
                                        double previousIterationPerformance = hydraPerformance.get(iter-1);
                                        int percentImprovement = (int) (100 * (previousIterationPerformance - performance) / previousIterationPerformance);
                                        if(percentImprovement < 0)
                                        {
                                            switch (options.portfolioEvaluation)
                                            {
                                                case VBS:
                                                    log.warn("VBS performance has got worse! This must be a bug!");
                                                case ZILLA:
                                                    log.warn("Zilla performance has got worse! Perhaps unpredictive configuration were added to portfolio.");
                                            }
                                        }
                                        log.info("Test performance runtime reduction of {}% over previous iteration", percentImprovement);
                                    }
                                } 
                            } 
                         } 
                         MyRunnable myRunnable = new MyRunnable(); 
                         Thread myThread = new Thread(myRunnable);
                         myThread.start(); 
						*/
						

					}
					
				}
				
				/****** Performance Report ******/
				log.info("*************Performance Summary*************");
				
			    if (bestSingleConfiguration != null)
			    {
			        log.info("Best single configuration: {}", bestSingleConfiguration.getFormattedParameterString(ParameterStringFormat.NODB_SYNTAX));
			    }
			    log.info("Final portfolio configurations");
			    int configCount = 1;
				for(ParameterConfiguration config : portfolio)
				{
					log.info("{}. Portfolio Configuration :  {}" ,  configCount, config.getFormattedParameterString(ParameterStringFormat.NODB_SYNTAX));
					configCount++;
				}
				
                /* Hydra Performance */
				log.info("Hydra performance by iteration:");
				String header = String.format("%-15s%-15s", "Iteration", "Performance");
				log.info(header);
				log.info("{}", StringUtils.repeat("-", 30));
				Set<Integer> sortedSet = new TreeSet<Integer>(hydraPerformance.keySet());
				for (int iter : sortedSet)
				{
					String line = String.format("%-15s%-15s", iter, hydraPerformance.get(iter));
					log.info(line);
				}
				log.info("{}", StringUtils.repeat("-", 30));
				
                //log.info("Best individual configuration performance: ", bestSingleConfigurationPerformance);
				
			}finally
			{
				myTAE.notifyShutdown();
				smacExecutionsTAE.notifyShutdown();
			}
			
			return AEATKReturnValues.SUCCESS;
			
		} catch(Throwable t)
		{
			System.out.flush();
			System.err.flush();
			
			System.err.println("Error occured running HYDRA ( " + t.getClass().getSimpleName() + " : "+ t.getMessage() +  ")\nError Log: " + logLocation);
			System.err.flush();
			
				if(log != null)
				{
					
					log.error( "Message: {}",t.getMessage());
					
					
					if(t instanceof ParameterException)
					{
						log.info("Don't forget that some options are set by default from files in ~/.aclib/");
						log.info("Exception stack trace", t);
						
						
					} else if(t instanceof TargetAlgorithmAbortException)
					{
						log.error("A serious problem occured during target algorithm execution and we are aborting execution ",t );
						
						
						log.error("We tried to call the target algorithm wrapper, but this call failed.");
						log.error("The problem is (most likely) somewhere in the wrapper.");
						log.error("There is also possibly additional error information above (in this log)");
						log.error("The easiest way to debug this problem is to manually execute the call we tried and see why it did not return the correct result");
						log.error("The required syntax is something like \"Final Result for ParamILS: x,x,x,x,x\".);");
						log.error("Specifically the regex we are matching is {}", CommandLineAlgorithmRun.AUTOMATIC_CONFIGURATOR_RESULT_REGEX);
					}	else
					{
						log.info("Maybe try running in DEBUG mode if you are missing information");
						
						log.error( "Exception:{}", t.getClass().getCanonicalName());
						StringWriter sWriter = new StringWriter();
						PrintWriter writer = new PrintWriter(sWriter);
						t.printStackTrace(writer);
						log.error( "StackTrace:{}",sWriter.toString());
					}
						
					log.info("Exiting Hydra with failure. Log: " + logLocation);
					log.info("For a list of available commands use:  --help");
					log.info("Please see above for the available options. Further information is available in the following documents:");
					log.info("- The Quickstart Guide (doc/quickstart.pdf) gives a simple example for getting up and running.");
					log.info("- The Manual (doc/manual.pdf) contains detailed information on file format semantics.");

					
					
					t = t.getCause();
				} else
				{
					if(t instanceof ParameterException )
					{
						
						System.err.println(t.getMessage());
						t.printStackTrace();
					} else
					{
						t.printStackTrace();
					}
					
				}
		
				
				if(t instanceof ParameterException)
				{
					return AEATKReturnValues.PARAMETER_EXCEPTION;
				}
				
				if(t instanceof StateSerializationException)
				{
					return AEATKReturnValues.SERIALIZATION_EXCEPTION;
				}
				
				if(t instanceof TrajectoryDivergenceException)
				{
					return AEATKReturnValues.TRAJECTORY_DIVERGENCE;
				}
				
				return AEATKReturnValues.OTHER_EXCEPTION;
		}
		
		
	}

	/* Convert Zilla algorithm result to Hydra result */
    private static Pair<AlgorithmRunConfiguration, AlgorithmRunResult> convertZillaResult(
            ca.ubc.cs.beta.zilla.utils.Pair<Algorithm, ProblemInstance> result,
            AlgorithmExecutionResult algorithmExecutionResult) 
    {
        JSONConverter<ParameterConfiguration> configConv = new JSONConverter<ParameterConfiguration>(){};

        ParameterConfiguration paramConfig = configConv.getObject(result.getFirst().getName());
        
        algorithmExecutionResult.getCutoff();
        double quality = algorithmExecutionResult.getQuality();
        RunStatus runStatus = algorithmExecutionResult.getResult();
        double runtime = algorithmExecutionResult.getRuntime();
        long seed = algorithmExecutionResult.getSeed();
        
        AlgorithmRunConfiguration runConfig = new AlgorithmRunConfiguration(new ProblemInstanceSeedPair(ZillaPortfolioEvaluator.convertZillatoAEATK(result.getSecond()),seed),paramConfig,execConfig);
        
        AlgorithmRunResult runResult = new ExistingAlgorithmRunResult(runConfig, runStatus, runtime,  -1.0, quality, seed);
        
        return new Pair(runConfig, runResult);
    }


    private static String runGroupName = "DEFAULT";
	
	/**
	 * Parsers Command Line Arguments and returns a options object
	 * @param args
	 * @return
	 */
	private static HydraOptions parseCLIOptions(String[] args) throws ParameterException, IOException
	{
		//DO NOT LOG UNTIL AFTER WE PARSE CONFIG OBJECT
	    
	    
	    /* What about hydra parameters???*/
		
	    HydraOptions options = new HydraOptions();
	   
		taeOptions = options.smacTAEOptions.getAvailableTargetAlgorithmEvaluators();
		//taeOptions.
		
		//System.out.println(JCommander.DEBUG_PROPERTY);
		
		//JCommander.DEBUG_PROPERTY = "dfdf";
		
		JCommander jcom = JCommanderHelper.getJCommanderAndCheckForHelp(args, options, taeOptions);
		
		jcom.setProgramName("hydra");
		
		try {
			try {
				try 
				{
				    args = processScenarioStateRestore(args);
				    jcom.parse(args);
				} finally
				{
					runGroupName = options.smacOptions.runGroupOptions.getFailbackRunGroup();
				}
				
				

				if(options.smacOptions.adaptiveCapping == null)
				{
					switch(options.smacOptions.scenarioConfig._runObj)
					{
					case RUNTIME:
						options.smacOptions.adaptiveCapping = true;
						break;
						
					case QUALITY:
						options.smacOptions.adaptiveCapping = false;
						break;
						
					default:
						//You need to add something new here
						throw new IllegalStateException("Not sure what to default too");
					}
				}
				
				if(options.smacOptions.randomForestOptions.logModel == null)
				{
					switch(options.smacOptions.scenarioConfig._runObj)
					{
					case RUNTIME:
						options.smacOptions.randomForestOptions.logModel = true;
						break;
					case QUALITY:
						options.smacOptions.randomForestOptions.logModel = false;
					}
				}

				
				
				//TODO: CLI TAE is somehwo null
				runGroupName = options.smacOptions.getRunGroupName(taeOptions.values());
				runGroupName = options.runGroupName;
				//fOutputDir = new File(options.smacOptions.scenarioConfig.outputDirectory);
				
				
				/*
				 * Build the Serializer object used in the model 
				 */
				outputDir = options.getOutputDirectory();
			
				File outputDirFile = new File(outputDir);
				
				
				if(!outputDirFile.exists())
				{
					outputDirFile.mkdirs();
					//Check again to ensure there isn't a race condition
					if(!outputDirFile.exists())
					{
						throw new ParameterException("Could not create all folders necessary for output directory: " + outputDir);
					}
					
				}
				
				
			} finally
			{
				
				options.smacOptions.logOptions.initializeLogging(outputDir, options.smacOptions.seedOptions.numRun);
				HydraExecutor.logLocation = options.smacOptions.logOptions.getLogLocation(outputDir,options.smacOptions.seedOptions.numRun);
				
				log = LoggerFactory.getLogger(HydraExecutor.class);
				
				exception = MarkerFactory.getMarker("EXCEPTION");
				stackTrace = MarkerFactory.getMarker("STACKTRACE");
				
				VersionTracker.setClassLoader(SPIClassLoaderHelper.getClassLoader());
				VersionTracker.logVersions();
				
				for(String name : jcom.getParameterFilesToRead())
				{
					log.info("Parsing (default) options from file: {} ", name);
				}
				
				
				log.info("Logging to {}." , outputDir);
			}
			
			log.trace("Command Line Options Parsed");
			
			//log.info("TAE selected was: {}",options.smacTAEOptions.targetAlgorithmEvaluator);
			
	
			
			JCommanderHelper.logCallString(args, HydraExecutor.class);
			

			 if(log.isDebugEnabled())
			 {
				Map<String, String> env = new TreeMap<String, String>(System.getenv());
					
				StringBuilder sb = new StringBuilder();
				 for (String envName : env.keySet()) {
					 sb.append(envName).append("=").append(env.get(envName)).append("\n");
					 
			           
			        }
				
				
					 
				 log.debug("==========Enviroment Variables===========\n{}", sb.toString());
				 
				 
				 Map<Object,Object > props = new TreeMap<Object, Object>(System.getProperties());
				 sb = new StringBuilder();
				 for (Entry<Object, Object> ent : props.entrySet())
				 {
					 
					 sb.append(ent.getKey().toString()).append("=").append(ent.getValue().toString()).append("\n");
					 
			           
			     }
				
				 String hostname = "[UNABLE TO DETERMINE HOSTNAME]";
				try {
					hostname = InetAddress.getLocalHost().getHostName();
				} catch(UnknownHostException e)
				{ //If this fails it's okay we just use it to output to the log
					
				}
				
				log.debug("Hostname:{}", hostname);
				log.debug("==========System Properties==============\n{}", sb.toString() );
			 }
			
			JCommanderHelper.logConfiguration(jcom);
			pool = options.seedOptions.getSeedableRandomPool();
			

			TrainTestInstances tti = options.smacOptions.getTrainingAndTestProblemInstances(pool, new SeedableRandomPool(options.smacOptions.validationSeed + options.smacOptions.seedOptions.seedOffset,pool.getInitialSeeds()));
			
			/* How are instance seeds generated */
			trainingILWS = tti.getTrainingInstances();
			testingILWS = tti.getTestInstances();
		
			try {
				//We don't handle this more gracefully because this seems like a super rare incident.
				if(ManagementFactory.getThreadMXBean().isThreadCpuTimeEnabled())
				{
					log.debug("JVM Supports CPU Timing Measurements");
				} else
				{
					log.warn("This Java Virtual Machine has CPU Time Measurements disabled, tunerTimeout will not contain any SMAC Execution Time.");
				}
			} catch(UnsupportedOperationException e)
			{
				log.warn("This Java Virtual Machine does not support CPU Time Measurements, tunerTimeout will not contain any SMAC Execution Time Information (http://docs.oracle.com/javase/1.5.0/docs/api/java/lang/management/ThreadMXBean.html#setThreadCpuTimeEnabled(boolean))");
			}
			
			if(options.smacOptions.seedOptions.numRun + options.smacOptions.seedOptions.seedOffset < 0)
			{
				log.warn("NumRun {} plus Seed Offset {} should be positive, things may not seed correctly",options.smacOptions.seedOptions.numRun, options.smacOptions.seedOptions.seedOffset );
			}
			
		    if(options.numConfigsPerIter > options.numParallelRuns)
		    {
		        log.warn("Num incumbent configurations to add to portfolio {} greater than number of SMAC runs {}. Taking all SMAC runs as new incumbents...",options.numConfigsPerIter,options.numParallelRuns);    
		        options.numConfigsPerIter=options.numParallelRuns;
		    }
			
			
			return options;
		} catch(IOException e)
		{
			throw e;
			
		} catch(ParameterException e)
		{
			throw e;
		}
	}
	
	private static String[] processScenarioStateRestore(String[] args) {
		return StateFactoryOptions.processScenarioStateRestore(args);
	
	}
		


}
 