package ca.ubc.cs.beta.hydra.incumbent;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import ca.ubc.cs.beta.aeatk.algorithmrunresult.AlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration.ParameterStringFormat;
import ca.ubc.cs.beta.zilla.algorithmexecution.AlgorithmExecutionResult;
import ca.ubc.cs.beta.zilla.algorithmexecution.managers.IAlgorithmExecutionManager;
import ca.ubc.cs.beta.zilla.algorithmexecution.managers.LookupAlgorithmExecutionManager;
import ca.ubc.cs.beta.zilla.algorithmexecution.solvingpolicy.VBSSolvingPolicy;
import ca.ubc.cs.beta.zilla.base.algorithms.Algorithm;
import ca.ubc.cs.beta.zilla.base.probleminstances.ProblemInstance;
import ca.ubc.cs.beta.zilla.evaluators.solvingpolicyselector.ISolvingPolicySelectorEvaluator;
import ca.ubc.cs.beta.zilla.evaluators.solvingpolicyselector.TestInstancesSolvingPolicySelectorEvaluator;
import ca.ubc.cs.beta.zilla.optimization.base.Solution;
import ca.ubc.cs.beta.zilla.optimization.optimizers.setfunction.ISetFunction;
import ca.ubc.cs.beta.zilla.optimization.optimizers.setfunction.ISetFunctionMinimizer;
import ca.ubc.cs.beta.zilla.optimization.optimizers.setfunction.ZillaLocalSearchSetFunctionMinimizer;
import ca.ubc.cs.beta.zilla.performance.aggregatemetrics.IAggregatePerformanceMetric;
import ca.ubc.cs.beta.zilla.performance.metrics.IPerformanceMetric;
import ca.ubc.cs.beta.zilla.selection.solvingpolicyselector.ConstantSolvingPolicySelector;
import ca.ubc.cs.beta.zilla.selection.solvingpolicyselector.ISolvingPolicySelector;
import ca.ubc.cs.beta.zilla.utils.CollectionsHelper;
import ca.ubc.cs.beta.zilla.utils.Pair;

public class AllRunsIncumbentGroupsEvaluator implements IIncumbentEvaluator {
	
	private Set<ProblemInstance> fInstances;
	private IPerformanceMetric fPerformanceMetric;
	private IAggregatePerformanceMetric fAggregatePerformanceMetric;
	private double fCutoff;
	
	private final int fLengthEpoch;
	private final int fNumEpoch;
	private final double fBadAcceptProb;
	private final Random fRandom;

	public AllRunsIncumbentGroupsEvaluator(Set<ProblemInstance> instances, IPerformanceMetric performanceMetric, IAggregatePerformanceMetric aAggregatePerformanceMetric, double aCutoff, int aLengthEpoch, int aNumEpoch, double aBadAcceptProb, Random aRandomGenerator)
	{
		fInstances = instances;
		fPerformanceMetric = performanceMetric;
		fAggregatePerformanceMetric = aAggregatePerformanceMetric;
		fCutoff = aCutoff;
		fLengthEpoch = aLengthEpoch;
		fNumEpoch = aNumEpoch;
		fBadAcceptProb = aBadAcceptProb;
		fRandom = aRandomGenerator;
	}
	
	@Override
	public List<ParameterConfiguration> getKBestIncumbents(
			List<AlgorithmRunResult> runResults,
			List<ParameterConfiguration> incumbents, int k) 
	{
		
		Map<Pair<Algorithm, ProblemInstance>, AlgorithmExecutionResult> executions = new HashMap<Pair<Algorithm, ProblemInstance>, AlgorithmExecutionResult>();
		IAlgorithmExecutionManager algorithmExecutionManager = new LookupAlgorithmExecutionManager(executions);
		
		final ISolvingPolicySelectorEvaluator solvingPolicySelectorEvaluator = new TestInstancesSolvingPolicySelectorEvaluator(
				algorithmExecutionManager, fInstances, fPerformanceMetric,
				fAggregatePerformanceMetric);
		
		Map<Algorithm,ParameterConfiguration> configAlgorithmMap = new HashMap<Algorithm,ParameterConfiguration>();
		Set<Algorithm> aSolvers = new HashSet<Algorithm>();
		for(ParameterConfiguration incumbent : incumbents)
		{
			Algorithm alg = new Algorithm(""+incumbent.getFormattedParameterString(ParameterStringFormat.NODB_SYNTAX));
			aSolvers.add(alg);
			configAlgorithmMap.put(alg,incumbent);
		}
		
		final ArrayList<Algorithm> solvers = new ArrayList<Algorithm>(aSolvers);
		
		ISetFunction objFunction = new ISetFunction() {

			@Override
			public double evaluate(final boolean[] aArgument) {

				Set<Algorithm> selectedAlgorithms = new HashSet<Algorithm>(
						CollectionsHelper.select(new ArrayList<Algorithm>(
								solvers), aArgument));

				ISolvingPolicySelector vbsSelector = new ConstantSolvingPolicySelector(
						new VBSSolvingPolicy(selectedAlgorithms,
								fPerformanceMetric, fCutoff));

				return solvingPolicySelectorEvaluator.evaluate(vbsSelector,
						fCutoff, 0);
			}

			@Override
			public int getSize() {
				return solvers.size();
			}

			@Override
			public int getMinimumSize() {
				return 1;
			}
		};
		ISetFunctionMinimizer localSearchFunction = new ZillaLocalSearchSetFunctionMinimizer(fLengthEpoch, fNumEpoch, fBadAcceptProb, fRandom);
		Solution<boolean[]> solution = localSearchFunction.minimize(objFunction);
		Collection<Algorithm> algorithms =  new ArrayList<Algorithm>(new HashSet<Algorithm>(
				CollectionsHelper.select(new ArrayList<Algorithm>(solvers),solution.getSolution())));
		
		
		List<ParameterConfiguration> bestkIncumbents = new ArrayList<ParameterConfiguration>();
		for (Algorithm alg: algorithms)
		{
			bestkIncumbents.add(configAlgorithmMap.get(alg));
		}
		return bestkIncumbents;
		
	}

}
