package ca.ubc.cs.beta.hydra.json;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import ca.ubc.cs.beta.hydra.builders.SMACInput;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ca.ubc.cs.beta.aeatk.algorithmrunresult.AlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.json.serializers.JsonDeserializerHelper;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;

import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.JsonToken;
import com.fasterxml.jackson.databind.DeserializationContext;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.deser.std.StdDeserializer;
import com.fasterxml.jackson.databind.ser.std.StdSerializer;

public class SMACInputJson {
    
    public static final String SI_ID = "@si-id";
    
    public static final String SI_PORTFOLIO = "si-portfolio";
    public static final String SI_RUNS = "si-runs";
    public static final String SI_ARGS = "si_args";

    
    public static class SMACInputDeserializer extends StdDeserializer<SMACInput>
    {
        private static final long serialVersionUID = 366966499180127827L;
        private final Logger log = LoggerFactory.getLogger(getClass());
        protected SMACInputDeserializer() {
            super(SMACInput.class);
        }

        @Override
        public SMACInput deserialize(JsonParser jp,DeserializationContext ctxt) throws IOException,JsonProcessingException 
        {
            //log.info("Deserialiing SMACInput object");
            
//            if(jp.getCurrentToken()==JsonToken.START_OBJECT)
//            {
//                jp.nextToken();
//            }
            
            ArrayList<String> args = new ArrayList<String>();
            String[] array_args = null;
            List<ParameterConfiguration> portfolio = new ArrayList<ParameterConfiguration>();
            List<AlgorithmRunResult> runs = new ArrayList<AlgorithmRunResult>();
            
            while(jp.nextToken() != JsonToken.END_OBJECT)
            {
                if(jp.getCurrentName() == null)
                {
                    continue;
                }
                
                switch(jp.getCurrentName())
                {
                    case SI_ARGS:
                        jp.nextToken();

                        while (jp.nextToken() != JsonToken.END_ARRAY) 
                        {
                            args.add(jp.getValueAsString());
                        }

                        array_args = new String[args.size()];
                        args.toArray(array_args);
                        break;
                    case SI_RUNS:
                        jp.nextToken();

                        while (jp.nextToken() != JsonToken.END_ARRAY) 
                        {
                            runs.add((AlgorithmRunResult) JsonDeserializerHelper.getDeserializedVersion(jp, ctxt, AlgorithmRunResult.class));
                        }
                        break;
                    case SI_PORTFOLIO:
                        jp.nextToken();

                        while (jp.nextToken() != JsonToken.END_ARRAY) 
                        {
                            portfolio.add((ParameterConfiguration) JsonDeserializerHelper.getDeserializedVersion(jp, ctxt, ParameterConfiguration.class));
                        }
                        break;
                }
                
            }
            return new SMACInput(array_args, runs, portfolio);
        }
        
    
    }
    
    public static class SMACInputSerializer extends StdSerializer<SMACInput>
    {
        private final Logger log = LoggerFactory.getLogger(getClass());
        protected SMACInputSerializer() 
        { 
            super(SMACInput.class);
        }

        private final ConcurrentHashMap<SMACInput, Integer> map = new ConcurrentHashMap<SMACInput, Integer>();
        
        private final AtomicInteger idMap = new AtomicInteger(1);
        
        @Override
        public void serialize(SMACInput input, JsonGenerator jgen,SerializerProvider provider) throws IOException,JsonGenerationException 
        {
            //log.info("Serialiing SMACInput object");
            
            jgen.writeStartObject();
            
            boolean firstWrite = (map.putIfAbsent(input,idMap.incrementAndGet()) == null);     
            
            if(firstWrite)
            {
                /* Write SMAC arguments */
                
                jgen.writeFieldName(SI_ARGS); // "myAction":
                jgen.writeStartArray(); // [
                for (String arg : input.fSMACArgs)
                        {
                            jgen.writeString(arg);
                        }
                jgen.writeEndArray(); // ]
                
                /* Write Portfolio */
                
                jgen.writeFieldName(SI_PORTFOLIO); // "myAction":
                jgen.writeStartArray(); // [
                for (ParameterConfiguration pc : input.fPortfolio)
                        {
                            jgen.writeObject(pc);
                        }
                jgen.writeEndArray(); // ]
                
                /* Write RunResults */
                
                jgen.writeFieldName(SI_RUNS); // "myAction":
                jgen.writeStartArray(); // [
                for (AlgorithmRunResult run : input.fRuns)
                        {
                            jgen.writeObject(run);
                        }
                jgen.writeEndArray(); // ]
            } 
            
            jgen.writeEndObject();
            
        }
        
    }

}
