package ca.ubc.cs.beta.hydra.json;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import ca.ubc.cs.beta.hydra.builders.SMACResult;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ca.ubc.cs.beta.aeatk.algorithmrunresult.AlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.json.serializers.JsonDeserializerHelper;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;

import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.JsonToken;
import com.fasterxml.jackson.databind.DeserializationContext;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.deser.std.StdDeserializer;
import com.fasterxml.jackson.databind.ser.std.StdSerializer;

public class SMACResultJson {
    
    public static final String SI_INCUMBENT = "si-incumbent";
    public static final String SI_RUNS = "si-runs";
    public static final String SI_TERMINATION = "si-termination";

    public static class SMACResultDeserializer extends StdDeserializer<SMACResult>
    {
        private static final long serialVersionUID = 366966499180127827L;
        private final Logger log = LoggerFactory.getLogger(getClass());

        protected SMACResultDeserializer() {
            super(SMACResult.class);
        }

        @Override
        public SMACResult deserialize(JsonParser jp,DeserializationContext ctxt) throws IOException,JsonProcessingException 
        {
            //log.info("Deserialiing SMACResult object");
            
//            if(jp.getCurrentToken()==JsonToken.START_OBJECT)
//            {
//                jp.nextToken();
//            }
            
            String terminationReason = null;
            ParameterConfiguration incumbent = null;
            List<AlgorithmRunResult> runs = new ArrayList<AlgorithmRunResult>();
            
            while(jp.nextToken() != JsonToken.END_OBJECT)
            {
                if(jp.getCurrentName() == null)
                {
                    continue;
                }
                
                switch(jp.getCurrentName())
                {
                    case SI_TERMINATION:
                        jp.nextToken();
                        
                        terminationReason = jp.getValueAsString();
                        break;
                    case SI_RUNS:
                        jp.nextToken();

                        while (jp.nextToken() != JsonToken.END_ARRAY) 
                        {
                            runs.add((AlgorithmRunResult) JsonDeserializerHelper.getDeserializedVersion(jp, ctxt, AlgorithmRunResult.class));
                        }
                        break;
                    case SI_INCUMBENT:
                        jp.nextToken();
                        
                        incumbent = (ParameterConfiguration) JsonDeserializerHelper.getDeserializedVersion(jp, ctxt, ParameterConfiguration.class);
                        break;
                }
                
            }
            return new SMACResult(incumbent, runs, terminationReason);
        }


    }

    public static class SMACResultSerializer extends StdSerializer<SMACResult>
    {
        private final Logger log = LoggerFactory.getLogger(getClass());

        protected SMACResultSerializer() {
            super(SMACResult.class);
        }
        
        private final ConcurrentHashMap<SMACResult, Integer> map = new ConcurrentHashMap<SMACResult, Integer>();
        
        private final AtomicInteger idMap = new AtomicInteger(1);
        

        @Override
        public void serialize(SMACResult result, JsonGenerator jgen,SerializerProvider provider) throws IOException,JsonGenerationException 
        {
            //log.info("Serialiing SMACResult object");
            
            jgen.writeStartObject();
            
            boolean firstWrite = (map.putIfAbsent(result,idMap.incrementAndGet()) == null);     
            
            if(firstWrite)
            {
                /* Write SMAC termination condition*/
                
                jgen.writeStringField(SI_TERMINATION, result.fTerminationReason);
                
                /* Write SMAC incumbent */
                
                jgen.writeObjectField(SI_INCUMBENT, result.fIncumbent);
     
                /* Write RunResults */
                
                jgen.writeFieldName(SI_RUNS); // "myAction":
                jgen.writeStartArray(); // [
                for (AlgorithmRunResult run : result.fRuns)
                        {
                            jgen.writeObject(run);
                        }
                jgen.writeEndArray(); // ]
            } 
            
            jgen.writeEndObject();
            
        }

    }

}
