package ca.ubc.cs.beta.hydra.options;

import java.util.Map;

import ca.ubc.cs.beta.aeatk.options.AbstractOptions;

/**
 * Container options object that decorates a generic abstract options
 * with a TAE options map (to keep the populated map around after TAE options parsing, for instance).
 * @author afrechet
 * @param <O> - the type of main abstract options.
 */
public class TAEAbstractOptionsWrapper<O extends AbstractOptions> extends AbstractOptions{

    private final O fAbstractOptions;
    private final Map<String,AbstractOptions> fTAEOptions;
    
    /**
     * @param aAbstractOptions - abstract options to decorate.
     * @param aTAEOptions - TAE options to decorate options with.
     */
    public TAEAbstractOptionsWrapper(O aAbstractOptions, Map<String,AbstractOptions> aTAEOptions)
    {
        fAbstractOptions = aAbstractOptions;
        fTAEOptions = aTAEOptions;
    }
    
    /**
     * @return - the main options contained in the wrapper.
     */
    public O getOptions()
    {
        return fAbstractOptions;
    }
    
    /**
     * @return the TAE options map contained in this wrapper.
     */
    public Map<String,AbstractOptions> getTAEOptionsMap()
    {
        return fTAEOptions;
    }
    
}