package ca.ubc.cs.beta.hydra.portfoliohistory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ca.ubc.cs.beta.aeatk.algorithmexecutionconfiguration.AlgorithmExecutionConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunconfiguration.AlgorithmRunConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.AlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.misc.associatedvalue.Pair;
import ca.ubc.cs.beta.aeatk.objectives.OverallObjective;
import ca.ubc.cs.beta.aeatk.objectives.PortfolioObjective;
import ca.ubc.cs.beta.aeatk.objectives.RunObjective;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;
import ca.ubc.cs.beta.aeatk.probleminstance.InstanceListWithSeeds;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstanceSeedPair;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluator;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluatorCallback;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.functionality.portfolio.PortfolioRunKillingPolicy;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.functionality.portfolio.PortfolioTargetAlgorithmEvaluatorDecorator;

/**
 * VBSPAR10 Evaluator
 * @author Chris
 *
 */
public class VBSPortfolioEvaluator implements IPortfolioEvaluator {

    final private InstanceListWithSeeds fInstances;
    private TargetAlgorithmEvaluator fTAE;
    final private RunObjective fObj;
    
    private final Logger log = LoggerFactory.getLogger(getClass());
    final private double fCutoff;
	final private OverallObjective fOverallObjective;
	final private PortfolioObjective fPortfolioObjective;


    public VBSPortfolioEvaluator(InstanceListWithSeeds testingILWS, TargetAlgorithmEvaluator myTAE, RunObjective _runObj, OverallObjective overall, double cutoff, PortfolioObjective portfolioObjective) 
    {
        fInstances = testingILWS;
        fTAE = myTAE;
        fObj = _runObj;
        fCutoff = cutoff;
        fOverallObjective = overall;
        fPortfolioObjective = portfolioObjective;
    }

    @Override
    public double getPerformance(List<ParameterConfiguration> paramConfigs, AlgorithmExecutionConfiguration execConfig, int hydraIter) 
    {
        
        log.info("Getting portfolio performance");
        
        List<Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>> portfolioConfigs = new ArrayList<Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>>();
        
        for(ParameterConfiguration paramConfig : paramConfigs)
        {
            portfolioConfigs.add(new Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>(execConfig, paramConfig));
        }
        
        TargetAlgorithmEvaluator portfolioTAE = new PortfolioTargetAlgorithmEvaluatorDecorator(fTAE, portfolioConfigs, fObj, false, PortfolioRunKillingPolicy.SLOWERDIES, fPortfolioObjective);
        
        List<AlgorithmRunConfiguration> runConfigs = new ArrayList<AlgorithmRunConfiguration>();
        
        for(ProblemInstance problem : fInstances.getInstances())
        {
            /* Not sure if this is correct (instance seed pairs are correct). Might want to check cache */
            /* TODO: Can run out of seeds for an instance. What does this mean? */
            //long seed = fInstances.getSeedGen().getNextSeed(problem);
            long seed = 0;
            
            ProblemInstanceSeedPair pisp = new ProblemInstanceSeedPair(problem,seed);
            runConfigs.add(new AlgorithmRunConfiguration(pisp, paramConfigs.get(0), execConfig));
        }
        
        List<AlgorithmRunResult> runResults = portfolioTAE.evaluateRun(runConfigs);
        //portfolioTAE.evaluateRunsAsync(runConfigs, callback, observer);
        //portfolioTAE.notifyShutdown();
        
        
        Collection<Double> performances = new ArrayList<Double>();
		for(AlgorithmRunResult result : runResults){
            double instancePerformance = fObj.getObjective(result);
            performances.add(instancePerformance);
        }
		
		//portfolioTAE.close();
		
        return fOverallObjective.aggregate(performances, fCutoff);
        
        
    }
    
    

}
