package ca.ubc.cs.beta.hydra.portfoliohistory;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ca.ubc.cs.beta.zilla.algorithmexecution.AlgorithmExecutionResult;
import ca.ubc.cs.beta.zilla.algorithmexecution.managers.FromFileAlgorithmExecutionManager;
import ca.ubc.cs.beta.zilla.algorithmexecution.managers.IAlgorithmExecutionManager;
import ca.ubc.cs.beta.zilla.algorithmexecution.managers.LookupAlgorithmExecutionManager;
import ca.ubc.cs.beta.zilla.algorithmexecution.solvingpolicy.ISolvingPolicy;
import ca.ubc.cs.beta.zilla.base.algorithms.Algorithm;
import ca.ubc.cs.beta.zilla.base.probleminstances.ProblemInstance;
import ca.ubc.cs.beta.zilla.evaluators.algorithmexecutionpolicy.IAlgorithmExecutionPolicyEvaluator;
import ca.ubc.cs.beta.zilla.evaluators.algorithmexecutionpolicy.TestInstancesAlgorithmExecutionPolicyEvaluator;
import ca.ubc.cs.beta.zilla.evaluators.solvingpolicyselector.TestInstancesSolvingPolicySelectorEvaluator;
import ca.ubc.cs.beta.zilla.execution.options.metrics.PerformanceMetricOptions;
import ca.ubc.cs.beta.zilla.execution.options.metrics.PerformanceMetricOptions.PerformanceMetricChoice;
import ca.ubc.cs.beta.zilla.execution.options.metrics.QualityPerformanceMetric;
import ca.ubc.cs.beta.zilla.execution.options.zilla.HydraZillaBuilderOptions;
import ca.ubc.cs.beta.zilla.execution.options.zilla.ZillaBuilderOptions;
import ca.ubc.cs.beta.zilla.execution.selector.ZillaSelectorExecutor;
import ca.ubc.cs.beta.zilla.execution.vbsselection.InitialSolversParameters;
import ca.ubc.cs.beta.zilla.features.Feature;
import ca.ubc.cs.beta.zilla.features.manager.IFeatureManager;
import ca.ubc.cs.beta.zilla.modelling.IModel;
import ca.ubc.cs.beta.zilla.modelling.IModelFactory;
import ca.ubc.cs.beta.zilla.modelling.IModelParameters;
import ca.ubc.cs.beta.zilla.modelling.base.WeightedPredictionData;
import ca.ubc.cs.beta.zilla.modelling.constant.ConstantModel;
import ca.ubc.cs.beta.zilla.modelling.decorators.FeatureEnforcerModelFactoryDecorator;
import ca.ubc.cs.beta.zilla.modelling.solverpairs.ISolverPairsModelFactory;
import ca.ubc.cs.beta.zilla.modelling.solverpairs.ISolverPairsModelParameters;
import ca.ubc.cs.beta.zilla.optimization.optimizers.setfunction.ISetFunctionMinimizer;
import ca.ubc.cs.beta.zilla.optimization.optimizers.setfunction.NOPSetFunctionMinimizer;
import ca.ubc.cs.beta.zilla.performance.aggregatemetrics.GeometricMeanPerformanceMetric;
import ca.ubc.cs.beta.zilla.performance.aggregatemetrics.IAggregatePerformanceMetric;
import ca.ubc.cs.beta.zilla.performance.aggregatemetrics.SumAggregatePerformanceMetric;
import ca.ubc.cs.beta.zilla.performance.metrics.CompletionPerformanceMetric;
import ca.ubc.cs.beta.zilla.performance.metrics.IPerformanceMetric;
import ca.ubc.cs.beta.zilla.performance.metrics.PARkPerformanceMetric;
import ca.ubc.cs.beta.zilla.presolver.AllPreSolversRuntimesPreSolverScheduler;
import ca.ubc.cs.beta.zilla.presolver.GreedyPreSolverScheduler;
import ca.ubc.cs.beta.zilla.presolver.IPreSolverScheduler;
import ca.ubc.cs.beta.zilla.presolver.KPreSolverScheduler;
import ca.ubc.cs.beta.zilla.selection.algorithmselector.InstanceWideAlgorithmSelectorStrategy;
import ca.ubc.cs.beta.zilla.selection.algorithmselector.PerformanceBasedInstanceWideAlgorithmSelector;
import ca.ubc.cs.beta.zilla.selection.algorithmselector.WinnerTakeAllInstanceWideAlgorithmSelector;
import ca.ubc.cs.beta.zilla.selection.solvingpolicyselector.zilla.ZillaEgg;
import ca.ubc.cs.beta.zilla.selection.solvingpolicyselector.zilla.builders.ZillaBuilder;
import ca.ubc.cs.beta.zilla.selection.solvingpolicyselector.zilla.builders.stage.IZillaStageBuilder;
import ca.ubc.cs.beta.zilla.selection.solvingpolicyselector.zilla.builders.stage.basic.BasicZillaStageBuilder;
import ca.ubc.cs.beta.zilla.utils.Pair;
import ca.ubc.cs.beta.zilla.votingsystem.IVotingSystem;
import ca.ubc.cs.beta.zilla.votingsystem.MajorityVoting;

import org.mangosdk.spi.processor.Options;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.base.Predicate;

import ca.ubc.cs.beta.aeatk.algorithmexecutionconfiguration.AlgorithmExecutionConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunconfiguration.AlgorithmRunConfiguration;
import ca.ubc.cs.beta.aeatk.algorithmrunresult.AlgorithmRunResult;
import ca.ubc.cs.beta.aeatk.objectives.OverallObjective;
import ca.ubc.cs.beta.aeatk.objectives.RunObjective;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration;
import ca.ubc.cs.beta.aeatk.parameterconfigurationspace.ParameterConfiguration.ParameterStringFormat;
import ca.ubc.cs.beta.aeatk.probleminstance.InstanceListWithSeeds;
import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstanceSeedPair;
import ca.ubc.cs.beta.aeatk.random.SeedableRandomPool;
import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.TargetAlgorithmEvaluator;
//import ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance;
//import ca.ubc.cs.beta.aeatk.targetalgorithmevaluator.decorators.functionality.portfolio.PortfolioTargetAlgorithmEvaluatorDecorator;
//import ca.ubc.cs.beta.zilla.algorithmexecution.managers.TAEAlgorithmExecutionManager;
//import ca.ubc.cs.beta.zilla.algorithmexecution.solvingpolicy.ISolvingPolicy;
//import ca.ubc.cs.beta.zilla.algorithmexecution.solvingpolicy.VBSSolvingPolicy;
//import ca.ubc.cs.beta.zilla.base.algorithms.Algorithm;
//import ca.ubc.cs.beta.zilla.utils.Pair;

public class ZillaPortfolioEvaluator implements IPortfolioEvaluator {

    private double fCutoff;
    private double fPreSolvingCutoff;
    private IFeatureManager fCompleteFeatureManager;
    private IFeatureManager fCheapFeatureManager;
    private SeedableRandomPool fRandomPool;
    private ArrayList<ProblemInstance> fTrainingInstances;
    private ArrayList<ProblemInstance> fValidationInstances;
    private IPerformanceMetric fPerformanceMetric;

    private final Logger log = LoggerFactory.getLogger(getClass());
    private IAggregatePerformanceMetric fAggregatePerformanceMetric;
    private ISolverPairsModelFactory fSolverPairsModelFactory;
    private HydraZillaBuilderOptions fZillaBuilderOptions;
    private Predicate<Feature> fFeatureCompletionPredicate;
    private final String fOutputDirLocation;
    private String fRunGroupName;
    private TargetAlgorithmEvaluator fTAE;
    private RunObjective fObj;
    private Set<Algorithm> fPreSolvers;
    private IPreSolverScheduler fPreSolverScheduler;
    private Set<ISolvingPolicy> fPreSolvingPolicies;
	private ArrayList<ProblemInstance> fTestingInstances;


    public ZillaPortfolioEvaluator(double aCutoff, HydraZillaBuilderOptions zillaOptions, SeedableRandomPool aRandomPool, TargetAlgorithmEvaluator myTAE, RunObjective _runObj, String dir, List<ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance> trainingInstances, List<ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance> testingInstances, String featureFilename, OverallObjective interInstanceObj)
    {
        log.info("Creating zilla portfolio evaluator...");
        
        fOutputDirLocation = dir;

        fZillaBuilderOptions = zillaOptions;
        
        fTAE = myTAE;
        
        fObj = _runObj;

        fRandomPool = aRandomPool;

        fCutoff = aCutoff;
        fPreSolvingCutoff = zillaOptions.fPreSolverOptions.getPreSolvingCutoff(fCutoff);
        fCompleteFeatureManager = zillaOptions.getFeatureManager(fRandomPool.getRandom("CompleteFeatureManager").nextLong(), featureFilename);
        fCheapFeatureManager = zillaOptions.fCheapFeatureManagerOptions.getFeatureManager(fRandomPool.getRandom("CheapFeatureManager").nextLong());

        //Instantiate instances
        fTrainingInstances = new ArrayList<ProblemInstance>();
        for (ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance instance : trainingInstances)
        {
        	fTrainingInstances.add(convertAEATKtoZilla(instance));
        }
        fValidationInstances = fTrainingInstances;
        fTestingInstances = new ArrayList<ProblemInstance>();
        for (ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance instance : testingInstances)
        {
        	fTestingInstances.add(convertAEATKtoZilla(instance));
        }

        switch(fObj)
        {
        case RUNTIME:
        	switch(interInstanceObj)
        	{
        	case MEAN:
        		fPerformanceMetric = new PARkPerformanceMetric(1);
        		break;
        	case MEAN10:
        		fPerformanceMetric = new PARkPerformanceMetric(10);
        		break;
        	case MEAN1000:
        		fPerformanceMetric = new PARkPerformanceMetric(1000);
        		break;
        	}
        	break;
        case QUALITY:
        	fPerformanceMetric = new QualityPerformanceMetric();
        	break;
        }
        log.info("Using {} performance metric.",fPerformanceMetric.toString());

        fAggregatePerformanceMetric =  zillaOptions.fAggregatePerformanceMetricOptions.getAggregatePerformanceMetric();
        log.info("Using {} aggregate performance metric.",fAggregatePerformanceMetric.toString());

        fSolverPairsModelFactory = zillaOptions.fSolverPairsModelOptions.getSolverPairsModelFactory();

        fFeatureCompletionPredicate = fCompleteFeatureManager.getFeatureCompletionPredicate();


    }

    @Override
    public double getPerformance(List<ParameterConfiguration> paramConfigs,
            AlgorithmExecutionConfiguration execConfig, int hydraIter) 
    {
        try
        {
            log.info("Calculating Zilla Portfolio performance...");
            
            List<ca.ubc.cs.beta.aeatk.misc.associatedvalue.Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>> portfolioConfigs = new ArrayList<ca.ubc.cs.beta.aeatk.misc.associatedvalue.Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>>();
            
            Set<Algorithm> solvers = new HashSet<Algorithm>();
            
            for(ParameterConfiguration paramConfig : paramConfigs)
            {
                Algorithm alg = new Algorithm(""+paramConfig.getFormattedParameterString(ParameterStringFormat.NODB_SYNTAX));
                solvers.add(alg);
            }
            
            Map<Pair<Algorithm,ProblemInstance>,AlgorithmExecutionResult> executions = new HashMap<Pair<Algorithm,ProblemInstance>,AlgorithmExecutionResult>();
            
            List<AlgorithmRunConfiguration> runConfigs = new ArrayList<AlgorithmRunConfiguration>();
            
            
            Collection<ProblemInstance> instances = new HashSet<ProblemInstance>();
            instances.addAll(fTrainingInstances);
            instances.addAll(fValidationInstances);
            instances.addAll(fTestingInstances);

            //Generate runConfigs of all configuration in portfolio
            for(ProblemInstance problem : instances)
            {
                for(ParameterConfiguration paramConfig : paramConfigs)
                {
                    /* Not sure if this is correct (instance seed pairs are correct). Might want to check cache */
                    /* TODO: Can run out of seeds for an instance. What does this mean? */
                    //long seed = fInstances.getSeedGen().getNextSeed(problem);
                    long seed = 0;
                    
                    ProblemInstanceSeedPair pisp = new ProblemInstanceSeedPair(convertZillatoAEATK(problem),seed);
                    runConfigs.add(new AlgorithmRunConfiguration(pisp, paramConfig, execConfig));
                }
            }
            log.info("{} instances", instances.size());
            log.info("{} configurations", paramConfigs.size());
            log.info("Running {} instance-configs for zilla evaluation with cutoff {}", runConfigs.size(), fCutoff);
            
            List<AlgorithmRunResult> runResults = fTAE.evaluateRun(runConfigs);
            
            /* If there is a single configuration, no need to run zilla */
            if(solvers.size() == 1)
            {
                log.info("Only one solver. No solver selection performed...");
                double performance = 0;
                for(AlgorithmRunResult result : runResults)
                {
                    double instancePerformance = fObj.getObjective(result);
                    performance += instancePerformance;
                }
                return performance / runResults.size();
            }
                
            /* Gather performance of each */
            
            for (AlgorithmRunResult result : runResults)
            {
            	
                Algorithm alg = new Algorithm(""+result.getParameterConfiguration().getFormattedParameterString(ParameterStringFormat.NODB_SYNTAX));
                ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance instance = result.getProblemInstance();
                ProblemInstance conversion = convertAEATKtoZilla(instance);
                Pair<Algorithm,ProblemInstance> key = new Pair<Algorithm, ProblemInstance>(alg, conversion);
                AlgorithmExecutionResult value = new AlgorithmExecutionResult(result.getRuntime(), result.getRunStatus(), fCutoff, result.getResultSeed(), result.getQuality());
                //log.info("Result is {}", fPerformanceMetric.evaluate(value));
                executions.put(key, value);
            }
            
            
            IAlgorithmExecutionManager algorithmExecutionManager = new LookupAlgorithmExecutionManager(executions);
    
            //Set<Algorithm>fSolvers = aZillaBuilderOptions.fExecutionOptions.getSolvers(algorithmExecutionManager.getAlgorithms());
    
            List<Algorithm> algorithms = new ArrayList<Algorithm>();
    
            int iter = 0;
            Map<Algorithm, Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>> taeConfigsMap = new HashMap<Algorithm, Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>>();
            for(ParameterConfiguration paramConfig : paramConfigs)
            {
                Algorithm alg = new Algorithm(""+paramConfig.getFormattedParameterString(ParameterStringFormat.NODB_SYNTAX));
                Pair<AlgorithmExecutionConfiguration, ParameterConfiguration> config = new Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>(execConfig, paramConfig);
                taeConfigsMap.put(alg, config);
    
                iter++;
            }
    
    
            IVotingSystem<Algorithm> aVotingSystem = new MajorityVoting<Algorithm>();
    
            ISolverPairsModelParameters fSolverPairModelParameters = fZillaBuilderOptions.fSolverPairsModelOptions.getSolverPairsModelParameters(algorithmExecutionManager,
                    fCutoff,
                    fPerformanceMetric,
                    solvers,
                    aVotingSystem,
                    fRandomPool.getRandom("SolverPairsModel").nextLong());
    

            /**
             * Assign features to instances.
             */
            log.info("Using {} training and {} validation instances.",fTrainingInstances.size(),fValidationInstances.size());
            //Attribute features to all instances (only if feature time is less than feature computation cutoff time).
            log.info("Assigning features to all instances.");
//            Collection<ProblemInstance> aAllInstances = new ArrayList<ProblemInstance>(fTrainingInstances.size()+fValidationInstances.size());
//            aAllInstances.addAll(fTrainingInstances);
//            aAllInstances.addAll(fValidationInstances);
            
            int aCompleteFeatureInstanceCount = 0;
            int aIncCompleteFeatureInstanceCount = 0;
            
            
            for(ProblemInstance aInstance : instances)
            {   
                Feature aCompleteFeatures = fCompleteFeatureManager.getFeatureComputation(aInstance).getFeatures();
                
                
                
                if(fFeatureCompletionPredicate.apply(aCompleteFeatures))
                {
                    aCompleteFeatureInstanceCount++;
                    aInstance.setFeature(aCompleteFeatures);
                }
                else
                {
                    aIncCompleteFeatureInstanceCount++;
                    log.trace("Features for instance {} were incomplete, assigning only cheap features.",aInstance);
                    Feature cheapFeatures = fCheapFeatureManager.getFeatureComputation(aInstance).getFeatures();
                    aInstance.setFeature(cheapFeatures);
                }
            }
            log.info("{} instances with complete features, {} instances with incomplete features.",aCompleteFeatureInstanceCount,aIncCompleteFeatureInstanceCount);
        
            log.info("Complete features: {}", fCompleteFeatureManager.getFeatureDomain());
            
            /*
            * Instantiate Zilla evaluator.
            */
           log.info("Initializing solving policy selector evaluator...");
           TestInstancesSolvingPolicySelectorEvaluator fSolvingPolicySelectorEvaluator = new TestInstancesSolvingPolicySelectorEvaluator(algorithmExecutionManager, fTrainingInstances, fPerformanceMetric, fAggregatePerformanceMetric);
           TestInstancesSolvingPolicySelectorEvaluator fPreSolvingScenarioEvaluator = new TestInstancesSolvingPolicySelectorEvaluator(algorithmExecutionManager, fTrainingInstances, fPerformanceMetric, fAggregatePerformanceMetric /*new GeometricMeanPerformanceMetric()*/);
   
           
            
            
            /**
             * Build complete feature prediction model.
             */
            log.info("Building complete feature prediction model...");
            Set<String> aCheapFeatures = fCheapFeatureManager.getFeatureDomain();
            log.info("Using cheap features {}.",aCheapFeatures);
            //log.info("Using Complete features: {}", fCompleteFeatureManager.getFeatureDomain);
    
            //Generate training data
            Collection<WeightedPredictionData<Boolean>> aCompleteFeatureModelTrainingData = new ArrayList<WeightedPredictionData<Boolean>>();
            log.debug("Building training data for complete feature prediction model.");
            for(ProblemInstance aTrainingInstance : instances)
            {               
                boolean aFeatureComplete = fFeatureCompletionPredicate.apply(aTrainingInstance.getFeature());
                double aFeatureRuntime = 0.0;//fCompleteFeatureManager.getFeatureComputation(aTrainingInstance).getRuntime();
                aCompleteFeatureModelTrainingData.add(new WeightedPredictionData<Boolean>(aTrainingInstance,aFeatureComplete,aFeatureRuntime));
            }
    
            IModelFactory aCompleteFeatureModelFactory = fZillaBuilderOptions.fCompleteFeatureComputationModelOptions.getFactory();
            IModelParameters aCompleteFeatureModelParameters = fZillaBuilderOptions.fCompleteFeatureComputationModelOptions.getParameters(fRandomPool.getRandom("CompleteFeatureComputationModel").nextLong());
            log.info("Using model {} with parameters {}", fZillaBuilderOptions.fCompleteFeatureComputationModelOptions.getChoice(),aCompleteFeatureModelParameters.toString());
    
            IModel<Boolean> fCompleteFeatureModel;
    		if (aCheapFeatures.isEmpty()) {
    			log.warn("There are no cheap features. Features will be computed on every instance.");
    			fCompleteFeatureModel = new ConstantModel<Boolean>(Boolean.TRUE);
    		} else {
    			fCompleteFeatureModel = new FeatureEnforcerModelFactoryDecorator(
    					aCompleteFeatureModelFactory, aCheapFeatures).getModel(
    					aCompleteFeatureModelParameters,
    					aCompleteFeatureModelTrainingData);
    		}

    
            log.info("Instantiating backup solver selector...");
            InstanceWideAlgorithmSelectorStrategy fBackupSolverSelector = InstanceWideAlgorithmSelectorStrategy.constructEmptyListInstanceWideAlgorithmSelectorStrategy(
                    new WinnerTakeAllInstanceWideAlgorithmSelector(taeConfigsMap.keySet(), fTrainingInstances, algorithmExecutionManager, fCutoff, fPerformanceMetric, fAggregatePerformanceMetric),
                    new PerformanceBasedInstanceWideAlgorithmSelector(taeConfigsMap.keySet(), algorithmExecutionManager, fCutoff, fPerformanceMetric, fAggregatePerformanceMetric)
                    ); 
    
            /**
             * Instantiate the solver subset selector.
             */
            log.info("Initializing the solver subset selector...");
            ISetFunctionMinimizer aSolverSubsetSelector;
            if(taeConfigsMap.keySet().size()==2)
            {
                log.warn("There are only two possible solvers, no solver subset selection will be performed.");
                aSolverSubsetSelector = new NOPSetFunctionMinimizer();
            }
            else
            {
                log.info("Using procedure {}",fZillaBuilderOptions.fSolverSubsetSelectorOptions.fSolverSubsetSelectorChoice);
                aSolverSubsetSelector = fZillaBuilderOptions.fSolverSubsetSelectorOptions.getSolverSubsetSelector(fRandomPool.getRandom("SolverSubsetSelector"));
                //aSolverSubsetSelector.
            }
            log.info("...done!");
            
            
            
            /**
             * Generate pre-solver schedules.
             */
            IAlgorithmExecutionPolicyEvaluator aPreSolverEvaluator = new TestInstancesAlgorithmExecutionPolicyEvaluator(
                    algorithmExecutionManager,
                    fTrainingInstances,
                    new CompletionPerformanceMetric(),
                    new SumAggregatePerformanceMetric());
            
            fPreSolvers = fZillaBuilderOptions.fPreSolverOptions.getPreSolvers(solvers,
                    fPreSolvingCutoff,
                    aPreSolverEvaluator,
                    fRandomPool.getRandom("PreSolversSelection").nextLong());
            HashSet<Double> aPreSolvingRuntimes = new HashSet<Double>(fZillaBuilderOptions.fPreSolverOptions.getPreSolvingRuntimes(fPreSolvingCutoff));
            
            log.info("Generating pre-solver schedules from pre-solvers {} and pre-solving runtimes {}...",fPreSolvers,aPreSolvingRuntimes);

            //TODO Send back pre-solver scheduler choice to pre-solving options.
            switch(fZillaBuilderOptions.fPreSolverOptions.fPreSolverSchedulerChoice)
            {
                case ALL:
                    fPreSolverScheduler = new AllPreSolversRuntimesPreSolverScheduler(fPreSolvers, aPreSolvingRuntimes,fPreSolvingCutoff);
                    break;
                case KALL:
                    fPreSolverScheduler = new KPreSolverScheduler(fPreSolvers, aPreSolvingRuntimes, 2, true);
                    break;
                case GREEDY:
                    double minRuntime = fZillaBuilderOptions.fPreSolverOptions.fPreSolverSchedulerMinRuntime;
                    fPreSolverScheduler = new GreedyPreSolverScheduler(fPreSolvers, aPreSolvingRuntimes, minRuntime);
                    break;
                default:
                    throw new IllegalArgumentException("Unrecognized pre-solver scheduler choice "+fZillaBuilderOptions.fPreSolverOptions.fPreSolverSchedulerChoice);
            }
            
            fPreSolvingPolicies = fPreSolverScheduler.getPreSolverSchedules(fTrainingInstances,algorithmExecutionManager);
            
            log.info("Pre-solving policies:");
            for(ISolvingPolicy s : fPreSolvingPolicies)
            {
                log.info("{}",s.toString());
            }
            
            log.info("...done!");
            
             IZillaStageBuilder zillaStageBuilder = new BasicZillaStageBuilder(
                    algorithmExecutionManager,
                    fCutoff,
                    fCompleteFeatureManager,
                    fCompleteFeatureModel,
                    fPreSolvingCutoff,
                    fFeatureCompletionPredicate,
                    fBackupSolverSelector,
                    fSolverPairsModelFactory,
                    fSolverPairModelParameters,
                    fSolvingPolicySelectorEvaluator,
                    fPreSolvingScenarioEvaluator,
                    aSolverSubsetSelector);

            ZillaBuilder zilla = new ZillaBuilder(fPreSolvingPolicies, zillaStageBuilder, fPerformanceMetric, fZillaBuilderOptions.fSolverPairsModelOptions, fZillaBuilderOptions.fPerformFinalBuildingPhase);
    
            ZillaEgg aZillaEgg = zilla.build(fTrainingInstances, fValidationInstances, fRandomPool.getPoolSeed());
    
            /**
             * Test zilla.
             */
            log.info("==================================================================");
            log.info("Evaluating zilla on testing instances...");
    
            Set<ProblemInstance> aTestInstances = new HashSet<ProblemInstance>(fTestingInstances);
    
            log.info("{} test instances.",aTestInstances.size());
    
            Map<ProblemInstance,AlgorithmExecutionResult> aResults = ZillaSelectorExecutor.evaluateZilla(
                    aZillaEgg.hatch(fCompleteFeatureManager),
                    aTestInstances,
                    fCheapFeatureManager,
                    algorithmExecutionManager,
                    fCutoff,
                    0.0,
                    fRandomPool, 
                    log);
    
            log.info("...done!");
    
            log.info("Saving results to disk...");
            //Write run results to disk.
            File aReportFile = new File(fOutputDirLocation+File.separator + "Iteration" + hydraIter);
            aReportFile.mkdirs();
            aReportFile =  new File(fOutputDirLocation+File.separator + "Iteration" + hydraIter + File.separator + "ZillaResults.csv");
            log.info("Writing results to file {}.",aReportFile);
    
            String aSelector = "zilla";//+fRunGroupName;
    
            ZillaSelectorExecutor.writeResults(aResults, aSelector, aReportFile, log);
    
            /* Writing training and testing run results */
            File ResultsFile = new File(fOutputDirLocation+File.separator + "Iteration" + hydraIter + File.separator + "ConfigRunResults.csv");
            
            /* Testing */
            Map<ProblemInstance, Map<Algorithm, AlgorithmExecutionResult>> testingConfigurationRunResults = new HashMap<ProblemInstance,Map<Algorithm, AlgorithmExecutionResult>>();
            for(ProblemInstance testingInstance :  instances)
            {
                Map<Algorithm, AlgorithmExecutionResult> instanceExecutions = new HashMap<Algorithm, AlgorithmExecutionResult>();
                for (Algorithm alg : taeConfigsMap.keySet())
                {
                    AlgorithmExecutionResult result = executions.get(new Pair<Algorithm,ProblemInstance>(alg, testingInstance));
                    instanceExecutions.put(alg, result);
                }
                testingConfigurationRunResults.put(testingInstance, instanceExecutions);
            }
            
            ZillaSelectorExecutor.writeResults(testingConfigurationRunResults, taeConfigsMap, ResultsFile, log);
            
            double performance = 0;
    
            for(ProblemInstance prob : aResults.keySet())
            {
                double instancePerformance = fPerformanceMetric.evaluate(aResults.get(prob));
                performance += instancePerformance;
            }
    
            return performance / aTestInstances.size();
            
        } catch(Exception e)
        {
            e.printStackTrace();
            throw new RuntimeException(e);
        }
    
    }
        
    


		public static ProblemInstance convertAEATKtoZilla(
                ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance instance) {
            // TODO Auto-generated method stub
            String name = instance.getInstanceName();
            return new ProblemInstance(name);
        }
        
        public static ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance convertZillatoAEATK(
                ProblemInstance instance) {
            // TODO Auto-generated method stub
            String name = instance.getName();
            return new ca.ubc.cs.beta.aeatk.probleminstance.ProblemInstance(name);
        } 
    

    //    @Override
    //    public double getPerformance(List<ParameterConfiguration> paramConfigs,
    //            AlgorithmExecutionConfiguration execConfig) {
    //        // TODO Auto-generated method stub
    //        @Override
    //        public double getPerformance(List<ParameterConfiguration> paramConfigs, AlgorithmExecutionConfiguration execConfig) 
    //        {
    //           
    ////            List<Algorithm> algorithms = new ArrayList<Algorithm>();
    ////            
    ////            ISolvingPolicy vbs = new VBSSolvingPolicy(algorithms, fPerformanceMetric, fCutoff);
    ////            
    ////            Map<Algorithm, Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>> taeConfigsMap = new HashMap<Algorithm, Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>>();
    ////            
    ////            int iter = 0;
    ////            for(ParameterConfiguration paramConfig : paramConfigs)
    ////            {
    ////                Algorithm alg = new Algorithm("Config" + iter);
    ////                Pair<AlgorithmExecutionConfiguration, ParameterConfiguration> config = new Pair<AlgorithmExecutionConfiguration, ParameterConfiguration>(execConfig, paramConfig);
    ////                taeConfigsMap.put(alg, config);
    ////                
    ////                iter++;
    ////            }
    ////            portfolioTAE = new PortfolioTargetAlgorithmEvaluatorDecorator(aTAE, aPortfolio, fObj, false, aPortfolioRunKillingPolicy.)
    ////            
    ////            IAlgorithmExecutionManager aExecutionManager = new TAEAlgorithmExecutionManager(fTAE, taeConfigsMap);
    ////            
    ////            for(ProblemInstance problem : fInstances.getInstances())
    ////            {
    ////                /* Not sure if this is correct (instance seed pairs are correct). Might want to check cache */
    ////                long seed = fInstances.getSeedGen().getNextSeed(problem);
    ////                vbs.execute(aExecutionManager, (ProblemInstance) problem, fCutoff, seed);
    //            }
    //            
    //        }
    //    }



}
